from __future__ import annotations

import logging
import secrets
from dataclasses import dataclass
from typing import TYPE_CHECKING, Generic, TypeVar, cast

import httpx
import orjson

from .types.cache import (
    flight_list_schema,
    playback_track_schema,
)
from .types.json import AirportList, Find, FlightList, Playback
from .utils import (
    DEFAULT_HEADERS,
    Err,
    Ok,
    Result,
    dataclass_opts,
    get_current_timestamp,
    raise_missing_polars,
    to_flight_id_hex,
    to_unix_timestamp,
)

if TYPE_CHECKING:
    from typing import Annotated, Any, Literal

    import polars as pl

    from .types import (
        IntoFlightId,
        IntoTimestamp,
    )
    from .types.cache import (
        FlightListRecord,
        PlaybackTrackEMSRecord,
        PlaybackTrackRecord,
    )
    from .types.json import (
        AirportRequest,
        Authentication,
        FlightData,
        FlightListItem,
        FlightListRequest,
        PlaybackRequest,
        TrackData,
    )

logger = logging.getLogger(__name__)


def get_json_headers(*, device: str | None = None) -> dict[str, str]:
    headers = DEFAULT_HEADERS.copy()
    headers["fr24-device-id"] = device or f"web-{secrets.token_urlsafe(32)}"
    return headers


def with_auth(
    mut_request_data: dict[str, Any],
    auth: Authentication | None,
    headers: httpx.Headers,
) -> dict[str, Any]:
    """Adds authentication details to a request data dictionary."""
    if (
        auth is not None
        and (sub_key := auth.get("userData", {}).get("subscriptionKey"))
        is not None
    ):
        mut_request_data["token"] = sub_key
    else:
        mut_request_data["device"] = headers.get(
            "fr24-device-id", f"web-{secrets.token_urlsafe(32)}"
        )

    return mut_request_data


# NOTE: we intentionally use dataclass to store request data so we can
# serialise it to disk easily.


@dataclass(**dataclass_opts)
class FlightListParams:
    """
    Parameters to fetch metadata/history of flights for
    *either* a given aircraft registration or flight number.
    """

    # NOTE: consider moving away from internally tagged enum
    reg: str | None = None
    """Aircraft registration (e.g. `B-HUJ`)"""
    flight: str | None = None
    """Flight number (e.g. `CX8747`)"""
    page: int = 1
    """Page number"""
    limit: int = 10
    """Number of results per page - use `100` if authenticated."""
    timestamp: IntoTimestamp | Literal["now"] | None = "now"
    """Show flights with ATD before this Unix timestamp"""
    filter_by: Literal["historic"] | None = None
    """If `historic`, do not return scheduled or live flights"""

    def __post_init__(self) -> None:
        if self.reg is None and self.flight is None:
            raise ValueError(
                "expected one of `reg` or `flight` to be set, "
                "but both are `None`"
            )
        if self.reg is not None and self.flight is not None:
            raise ValueError(
                "expected only one of `reg` or `flight` to be set, "
                "but both are not `None`"
            )

    @property
    def kind(self) -> Literal["reg", "flight"]:
        return "reg" if self.reg is not None else "flight"

    @property
    def ident(self) -> str:
        return self.reg if self.reg is not None else self.flight  # type: ignore


async def flight_list(
    client: httpx.AsyncClient,
    params: FlightListParams,
    headers: httpx.Headers,
    auth: None | Authentication,
) -> Annotated[httpx.Response, FlightList]:
    """
    Query flight list data.

    To determine if there are more pages to query, check the response
    [.result.response.page.more][fr24.types.json.Page.more].

    Includes basic information such as status, O/D, scheduled/estimated/real
    times: see [fr24.types.json.FlightList][] for more details.

    :param client: HTTPX async client
    :param auth: Authentication data
    """
    timestamp = to_unix_timestamp(params.timestamp)
    if timestamp == "now":
        timestamp = get_current_timestamp()

    key, value = params.kind, params.ident

    request_data: FlightListRequest = {
        "query": value,
        "fetchBy": key,
        "page": params.page,
        "limit": params.limit,
    }
    if params.filter_by is not None:
        request_data["filterBy"] = params.filter_by
    if timestamp is not None:
        request_data["timestamp"] = timestamp

    request = httpx.Request(
        "GET",
        "https://api.flightradar24.com/common/v1/flight/list.json",
        headers=headers,
        params=with_auth(request_data, auth, headers),  # type: ignore
    )

    response = await client.send(request)
    return response


@dataclass(**dataclass_opts)
class AirportListParams:
    """
    Request data to fetch metadata/history of flights
    """

    airport: str
    """IATA airport code (e.g. `HKG`)"""
    mode: Literal["arrivals", "departures", "ground"]
    """arrivals, departures or on ground aircraft"""
    page: int = 1
    """Page number"""
    limit: int = 10
    """Number of results per page - use `100` if authenticated."""
    timestamp: IntoTimestamp | Literal["now"] | None = "now"
    """Show flights with STA before this timestamp"""


async def airport_list(
    client: httpx.AsyncClient,
    params: AirportListParams,
    headers: httpx.Headers,
    auth: None | Authentication,
) -> Annotated[httpx.Response, AirportList]:
    """
    Fetch aircraft arriving, departing or on ground at a given airport.

    Returns on ground/scheduled/estimated/real times: see
    [fr24.types.json.FlightListItem][] for more details.

    :param client: HTTPX async client
    :param auth: Authentication data
    :returns: the raw binary response, representing a JSON-encoded
        [fr24.types.json.FlightList][].
    """
    timestamp = to_unix_timestamp(params.timestamp)
    if timestamp == "now":
        timestamp = get_current_timestamp()

    request_data: AirportRequest = {
        "code": params.airport,
        "plugin[]": ["schedule"],
        "plugin-setting[schedule][mode]": params.mode,
        "page": params.page,
        "limit": params.limit,
    }

    if timestamp is not None:
        request_data["plugin-setting[schedule][timestamp]"] = timestamp

    request = httpx.Request(
        "GET",
        "https://api.flightradar24.com/common/v1/airport.json",
        headers=headers,
        params=with_auth(request_data, auth, headers),  # type: ignore
    )

    response = await client.send(request)
    return response


@dataclass(**dataclass_opts)
class PlaybackParams:
    """
    Request data to fetch historical track playback data for a given flight.
    """

    flight_id: IntoFlightId
    """fr24 flight id, represented in hex"""
    timestamp: IntoTimestamp | None = None
    """Actual time of departure (ATD) of the historic flight,
    Unix timestamp in seconds. Optional, but it is recommended to include it.
    """


async def playback(
    client: httpx.AsyncClient,
    params: PlaybackParams,
    headers: httpx.Headers,
    auth: None | Authentication,
) -> Annotated[httpx.Response, Playback]:
    """
    Fetch historical track playback data for a given flight.

    :param client: HTTPX async client
    :param auth: Authentication data
    """
    timestamp = to_unix_timestamp(params.timestamp)
    if timestamp == "now":
        timestamp = get_current_timestamp()

    request_data: PlaybackRequest = {
        "flightId": to_flight_id_hex(params.flight_id),
    }
    if timestamp is not None:
        request_data["timestamp"] = timestamp

    request = httpx.Request(
        "GET",
        "https://api.flightradar24.com/common/v1/flight-playback.json",
        headers=headers,
        params=with_auth(request_data, auth, headers),  # type: ignore
    )

    response = await client.send(request)
    return response


@dataclass(**dataclass_opts)
class FindParams:
    query: str
    """Airport, schedule (HKG-CDG), or aircraft."""
    limit: int = 50


async def find(
    client: httpx.AsyncClient,
    params: FindParams,
    headers: httpx.Headers,
    auth: None | Authentication,
) -> Annotated[httpx.Response, Find]:
    """General search."""
    request_data = {
        "query": params.query,
        "limit": params.limit,
    }
    request = httpx.Request(
        "GET",
        url="https://www.flightradar24.com/v1/search/web/find",
        headers=headers,
        params=with_auth(request_data, auth, headers),
    )
    response = await client.send(request)
    return response


#
# helpers
#

# workaround for py<3.12: https://docs.python.org/3/reference/compound_stmts.html#type-params
# just to silence mypy
_TypedDictT = TypeVar("_TypedDictT")


class _Parser(Generic[_TypedDictT]):
    @staticmethod
    def parse_json(
        response: Annotated[httpx.Response, _TypedDictT],
    ) -> Result[_TypedDictT, httpx.HTTPStatusError]:
        """Parses binary representation into a python object (typed dict)."""
        if response.is_success:
            return Ok(cast(_TypedDictT, orjson.loads(response.content)))
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            return Err(e)
        else:
            raise RuntimeError(
                "unexpected code path: response did not raise an error!"
            )


flight_list_parse = _Parser[FlightList].parse_json
airport_list_parse = _Parser[AirportList].parse_json
playback_parse = _Parser[Playback].parse_json
find_parse = _Parser[Find].parse_json


def playback_metadata_dict(flight: FlightData) -> dict[str, Any]:
    """
    Flatten and rename important variables in the flight metadata to a
    dictionary.
    """
    ident = flight["identification"]
    sta = flight["status"]["generic"]
    owner = flight["owner"]
    aircraft = flight["aircraft"]
    airline = flight["airline"]
    origin = flight["airport"]["origin"]
    dest = flight["airport"]["destination"]
    return {
        "flight_id": int(id_, 16)
        if isinstance(id_ := ident["id"], str)
        else id_,
        "callsign": ident["callsign"],
        "flight_number": ident["number"].get("default", None),
        "status_type": sta["status"]["type"],
        "status_text": sta["status"]["text"],
        "status_diverted": sta["status"]["diverted"],
        "status_time": int(time)
        if (time := sta["eventTime"]["utc"]) is not None
        else None,
        "model_code": aircraft["model"]["code"]
        if aircraft is not None
        else None,
        "icao24": int(aircraft["identification"]["modes"], 16)
        if aircraft is not None
        else None,
        "registration": aircraft["identification"]["registration"]
        if aircraft is not None
        else None,
        "owner": owner["code"]["icao"] if owner is not None else None,
        "airline": airline["code"]["icao"] if airline is not None else None,
        "origin": origin["code"]["icao"] if origin is not None else None,
        "destination": dest["code"]["icao"] if dest is not None else None,
        "median_delay": delay_med
        if (delay_med := flight["median"]["delay"]) is not None
        else None,
        "median_time": int(time_med)
        if (time_med := flight["median"]["timestamp"]) is not None
        else None,
    }


def playback_track_dict(point: TrackData) -> PlaybackTrackRecord:
    """
    Flatten and rename each variable in this observation into a new dictionary.

    !!! note
        The JSON response claims that `heading` is available, but ADS-B only
        transmits the [ground track](https://mode-s.org/decode/content/ads-b/4-surface-position.html#ground-track).
        [Heading](https://mode-s.org/decode/content/mode-s/7-ehs.html#heading-and-speed-report-bds-60)
        is only available in [EMS][fr24.types.json.EMS] data.

        We rename it to `track` to avoid confusion.
    """
    return {
        "timestamp": point["timestamp"],
        "latitude": point["latitude"],
        "longitude": point["longitude"],
        "altitude": point["altitude"]["feet"],
        "ground_speed": point["speed"]["kts"],
        "vertical_speed": point["verticalSpeed"]["fpm"],
        "track": point["heading"],
        "squawk": int(point["squawk"], base=8),
        "ems": playback_track_ems_dict(point),
    }


def playback_track_ems_dict(point: TrackData) -> PlaybackTrackEMSRecord | None:
    """
    If the Enhanced Mode-S data is available in this observation,
    flatten and rename each variable to a dictionary. Otherwise, return `None`.
    """
    if e := point["ems"]:
        return {
            "timestamp": e["ts"],
            "ias": e["ias"],
            "tas": e["tas"],
            "mach": e["mach"],
            "mcp": e["mcp"],
            "fms": e["fms"],
            "autopilot": e["autopilot"],
            "oat": e["oat"],
            "track": e["trueTrack"],
            "roll": e["rollAngle"],
            "qnh": e["qnh"],
            "wind_dir": e["windDir"],
            "wind_speed": e["windSpd"],
            "precision": e["precision"],
            "altitude_gps": e["altGPS"],
            "emergency": e["emergencyStatus"],
            "tcas_acas": e["tcasAcasDtatus"],
            "heading": e["heading"],
        }
    return None


def playback_df(data: Playback) -> pl.DataFrame:
    """
    Parse each [fr24.types.json.TrackData][] in the API response into a
    dataframe.

    If the response is empty, a warning is logged and an empty table is returned
    """
    try:
        import polars as pl
    except ImportError as exc:
        raise_missing_polars(exc)

    flight = data["result"]["response"]["data"]["flight"]
    if len(track := flight["track"]) == 0:
        logger.warning("no data in response, table will be empty")
    return pl.DataFrame(
        [playback_track_dict(point) for point in track],
        schema=playback_track_schema,
    )
    # NOTE: original implementation returns pl.DateTime instead of timestamp


def flight_list_dict(entry: FlightListItem) -> FlightListRecord:
    """
    Flatten a flight entry into dict, converting fr24 hex ids into integers.
    """
    orig = entry["airport"]["origin"]
    dest = entry["airport"]["destination"]
    icao24 = entry["aircraft"]["hex"]
    id_ = entry["identification"]["id"]
    stod = entry["time"]["scheduled"]["departure"]
    etod = entry["time"]["estimated"]["departure"]
    atod = entry["time"]["real"]["departure"]
    stoa = entry["time"]["scheduled"]["arrival"]
    etoa = entry["time"]["estimated"]["arrival"]
    atoa = entry["time"]["real"]["arrival"]
    return {
        "flight_id": int(id_, 16) if id_ is not None else None,
        "number": entry["identification"]["number"]["default"],
        "callsign": entry["identification"]["callsign"],
        "icao24": int(icao24, 16) if icao24 is not None else None,
        "registration": entry["aircraft"]["registration"],
        "typecode": entry["aircraft"]["model"]["code"],
        "origin": orig["code"]["icao"] if orig is not None else None,
        "destination": dest["code"]["icao"] if dest is not None else None,
        "status": entry["status"]["text"],
        "STOD": stod * 1000 if stod is not None else None,
        "ETOD": etod * 1000 if etod is not None else None,
        "ATOD": atod * 1000 if atod is not None else None,
        "STOA": stoa * 1000 if stoa is not None else None,
        "ETOA": etoa * 1000 if etoa is not None else None,
        "ATOA": atoa * 1000 if atoa is not None else None,
    }


def flight_list_df(data: FlightList) -> pl.DataFrame:
    """
    Parse each [fr24.types.json.FlightListItem][] in the API response
    into a polars dataframe.

    If the response is empty, a warning is logged and an empty table is returned
    """
    try:
        import polars as pl
    except ImportError as exc:
        raise_missing_polars(exc)

    if (flights := data["result"]["response"]["data"]) is None:
        logger.warning("no data in response, table will be empty")
        return pl.DataFrame(schema=flight_list_schema)
    return pl.DataFrame(
        [flight_list_dict(f) for f in flights],
        schema=flight_list_schema,
    )



from typing import Literal

LiveFeedFieldAuthenticated = Literal[
    "squawk", "vspeed", "airspace", "logo_id", "age"
]

LiveFeedField = Literal[
    "flight", "reg", "route", "type", LiveFeedFieldAuthenticated
]



# ruff: noqa
# mypy: disable-error-code="top-level-await, no-redef"
# %%
# --8<-- [start:script0]
import httpx
from fr24.grpc import top_flights
from fr24.proto.v1_pb2 import TopFlightsRequest, TopFlightsResponse
from fr24.proto import parse_data
from fr24.proto.headers import get_grpc_headers


async def top_flights_data() -> TopFlightsResponse:
    headers = httpx.Headers(get_grpc_headers(auth=None))
    async with httpx.AsyncClient() as client:
        message = TopFlightsRequest(limit=10)
        response = await top_flights(client, message, headers)
        return parse_data(response.content, TopFlightsResponse).unwrap()


data = await top_flights_data()
data
# --8<-- [end:script0]
# %%
"""
# --8<-- [start:output0]
scoreboard_list {
  flight_id: 911336284
  live_clicks: 1648
  total_clicks: 6664
  callsign: "LEE92"
  from_iata: "QKG"
  from_city: "Leeming"
  type: "HUNT"
  full_description: "Hawker Hunter T2"
}
scoreboard_list {
  flight_id: 911326556
  live_clicks: 1144
  total_clicks: 8848
  callsign: "SARAB"
  type: "HAWK"
  full_description: "British Aerospace Hawk Mk167"
}
scoreboard_list {
  flight_id: 911338539
  live_clicks: 904
  total_clicks: 1764
  flight_number: "TN80001"
  callsign: "TN080001"
  from_iata: "QCY"
...
  to_iata: "RMQ"
  to_city: "Taichung"
  type: "B738"
  full_description: "Boeing 737-8AL"
}
# --8<-- [end:output0]
"""



# ruff: noqa
# mypy: disable-error-code="top-level-await, no-redef"
# %%
# --8<-- [start:script0]
import httpx
from fr24.grpc import live_flights_status
from fr24.proto.v1_pb2 import (
    LiveFlightsStatusRequest,
    LiveFlightsStatusResponse,
)
from fr24.proto import parse_data
from fr24.proto.headers import get_grpc_headers


async def live_flights_status_data() -> LiveFlightsStatusResponse:
    headers = httpx.Headers(get_grpc_headers(auth=None))
    async with httpx.AsyncClient() as client:
        message = LiveFlightsStatusRequest(
            flight_ids_list=[0x35FBC363, 0x35FBF180]
        )
        response = await live_flights_status(client, message, headers)
        return parse_data(response.content, LiveFlightsStatusResponse).unwrap()


data = await live_flights_status_data()
data
# --8<-- [end:script0]
# %%
"""
# --8<-- [start:output0]
flights_map {
  flight_id: 905703808
  data {
    lat: 21.7237415
    lon: 114.917908
    status: LIVE
  }
}
flights_map {
  flight_id: 905692003
  data {
    lat: 21.9285736
    lon: 113.912445
    status: LIVE
  }
}
# --8<-- [end:output0]
"""



import pytest

from fr24 import FR24, FR24Cache
from fr24.service import FlightListResult, PlaybackFlightResult
from fr24.utils import to_flight_id


@pytest.fixture
async def flight_list_result(
    fr24: FR24,
) -> FlightListResult:
    REG = "B-LXA"
    result = await fr24.flight_list.fetch(reg=REG)
    return result


@pytest.fixture
async def playback_flight_result(
    fr24: FR24,
    flight_list_result: FlightListResult,
) -> PlaybackFlightResult:
    import polars as pl

    df = flight_list_result.to_polars()
    landed = df.filter(pl.col("status").str.starts_with("Landed"))

    assert landed.height > 0, "no landed flights found"

    flight_id = landed[0, "flight_id"]
    stod = int(landed[0, "ATOD"].timestamp())

    result = await fr24.playback_flight.fetch(
        flight_id=flight_id, timestamp=stod
    )
    return result


@pytest.mark.anyio
async def test_playback_flight_simple(
    playback_flight_result: PlaybackFlightResult,
) -> None:
    proto = playback_flight_result.to_proto()
    assert len(proto.flight_trail_list) > 10

    df = playback_flight_result.to_polars()
    assert df.height == 1
    assert df.width == 24


@pytest.mark.anyio
async def test_playback_flight_file_ops(
    playback_flight_result: PlaybackFlightResult, cache: FR24Cache
) -> None:
    flight_id = f"{to_flight_id(playback_flight_result.request.flight_id):0x}"
    timestamp = playback_flight_result.request.timestamp
    fp = (
        cache.path
        / "playback_flight"
        / f"{flight_id.upper()}_{timestamp}.parquet"
    )
    fp.parent.mkdir(parents=True, exist_ok=True)
    fp.unlink(missing_ok=True)

    playback_flight_result.write_table(cache)
    assert fp.exists(), (
        f"{fp} not in {list(f.name for f in fp.parent.glob('*'))}"
    )

    df_local = cache.playback_flight.scan_table(flight_id, timestamp).collect()
    assert df_local.equals(playback_flight_result.to_polars())

    fp.unlink()



import pytest

from fr24 import FR24, FR24Cache
from fr24.service import TopFlightsResult


@pytest.fixture
async def top_flights_result(
    fr24: FR24,
) -> TopFlightsResult:
    LIMIT = 10

    result = await fr24.top_flights.fetch(limit=LIMIT)
    return result


@pytest.mark.anyio
async def test_top_flights_fetch(top_flights_result: TopFlightsResult) -> None:
    num_flights_requested = top_flights_result.request.limit
    num_flights = len(top_flights_result.to_proto().scoreboard_list)
    assert num_flights_requested == num_flights

    df = top_flights_result.to_polars()
    assert df.height == num_flights
    assert df.width == 12


@pytest.mark.anyio
async def test_top_flights_file_ops(
    top_flights_result: TopFlightsResult, cache: FR24Cache
) -> None:
    timestamp = top_flights_result.timestamp
    ident = f"{timestamp}"
    fp = cache.path / "top_flights" / f"{ident}.parquet"
    fp.parent.mkdir(parents=True, exist_ok=True)
    fp.unlink(missing_ok=True)

    top_flights_result.write_table(cache)
    assert fp.exists(), (
        f"{fp} not in {list(f.name for f in fp.parent.glob('*'))}"
    )

    df_local = cache.top_flights.scan_table(ident).collect()
    assert df_local.equals(top_flights_result.to_polars())

    fp.unlink()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: fr24/proto/_live_trail.proto
# Protobuf Python Version: 6.33.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    0,
    '',
    'fr24/proto/_live_trail.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from fr24.proto import _common_pb2 as fr24_dot_proto_dot___common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1c\x66r24/proto/_live_trail.proto\x1a\x18\x66r24/proto/_common.proto\"%\n\x10LiveTrailRequest\x12\x11\n\tflight_id\x18\x01 \x01(\x07\"L\n\x11LiveTrailResponse\x12\x37\n\x12radar_records_list\x18\x01 \x03(\x0b\x32\x1b._common.RadarHistoryRecordb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'fr24.proto._live_trail_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_LIVETRAILREQUEST']._serialized_start=58
  _globals['_LIVETRAILREQUEST']._serialized_end=95
  _globals['_LIVETRAILRESPONSE']._serialized_start=97
  _globals['_LIVETRAILRESPONSE']._serialized_end=173
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import asyncio
import json
from datetime import datetime, timezone
from pathlib import Path
from typing import Iterator, Literal, TypeVar

import httpx
from textual import on
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import ScrollableContainer
from textual.widgets import DataTable, Footer, Header, Input, Label, Static

from fr24 import FR24
from fr24.tui.formatters import Time, fmt_aircraft, fmt_airport, fmt_status
from fr24.tui.widgets import AircraftWidget, AirportWidget, FlightWidget
from fr24.types import IntoTimestamp
from fr24.types.isqx import TimestampS
from fr24.types.json import FlightList, FlightListItem, is_schedule
from fr24.utils import UnwrapError, get_current_timestamp, to_unix_timestamp

T = TypeVar("T")


def flatten(*args: list[T]) -> Iterator[T]:
    for elt in args:
        yield from elt


class SearchBlock(Static):
    def compose(self) -> ComposeResult:
        yield Label("date")
        self.date_input = Input(id="date")
        self.date_input.value = f"{datetime.now(tz=timezone.utc):%Y-%m-%d}"
        yield self.date_input
        yield AircraftWidget(name="aircraft")
        yield FlightWidget(name="number")
        yield AirportWidget(id="departure", name="origin")
        yield AirportWidget(id="arrival", name="destination")


class FR24Tui(App[None]):
    CSS_PATH = "style.tcss"
    BINDINGS = [  # noqa: RUF012
        ("q", "quit", "Quit"),
        ("l", "login", "Log in"),
        ("r", "refresh", "Refresh"),
        ("/", "search", "Search"),
        ("s", "save", "Save"),
        ("c", "clear", "Clear"),
        Binding("escape", "escape", show=False),
    ]
    line_info: dict[str, str] = {}  # noqa: RUF012

    def compose(self) -> ComposeResult:
        self.fr24 = FR24()
        self.search_visible = True
        yield Header()
        yield Footer()
        yield SearchBlock()
        yield ScrollableContainer(DataTable())

    async def on_mount(self) -> None:
        await self.fr24.__aenter__()
        self.title = "FlightRadar24"
        table = self.query_one(DataTable)
        table.cursor_type = "row"
        table.add_columns(
            "date",
            "number",
            "callsign",
            "aircraft",
            "from",
            "to",
            "STD",
            "ATD",
            "STA",
            "status",
            "flightid",
        )

    async def on_unmount(self) -> None:
        await self.fr24.__aexit__(None, None, None)

    def action_search(self) -> None:
        self.search_visible = not self.search_visible
        if self.search_visible:
            self.query_one(SearchBlock).remove_class("hidden")
        else:
            self.query_one(SearchBlock).add_class("hidden")
            self.query_one(DataTable).focus()

    def action_clear(self) -> None:
        for widget in self.query(Input):
            if widget.id != "date":
                widget.value = ""

    async def action_escape(self) -> None:
        if not self.search_visible:
            await self.action_quit()

        self.search_visible = False
        self.query_one(SearchBlock).add_class("hidden")
        self.query_one(DataTable).focus()

    async def action_login(self) -> None:
        await self.fr24.login()
        if self.fr24.http.auth is not None:
            auth = self.fr24.http.auth
            identity = auth.get("user", {}).get("identity") or auth[
                "userData"
            ].get("identity")
            self.sub_title = (
                f"(authenticated: {identity})"
                if identity
                else "(authenticated)"
            )
            self.query_one(Header).add_class("authenticated")
            self.query_one(Footer).add_class("authenticated")

    async def on_data_table_row_selected(
        self, event: DataTable.RowSelected
    ) -> None:
        columns = [c.label.plain for c in event.data_table.columns.values()]
        self.line_info = dict(
            zip(columns, event.data_table.get_row(event.row_key))
        )

    async def action_save(self) -> None:
        if len(self.line_info) == 0:
            return
        date = self.line_info["date"] + " " + self.line_info["STD"]
        timestamp = to_unix_timestamp(date, format="%Y-%m-%d %H:%MZ")
        assert timestamp != "now"
        result = await self.fr24.playback.fetch(
            flight_id=self.line_info["flightid"],
            timestamp=timestamp,
        )
        result_dict = result.to_dict()
        filename = f"{self.line_info['flightid']}.json"
        self.notify(f"Saving to {filename}")
        Path(filename).write_text(json.dumps(result_dict, indent=2))

    @on(Input.Submitted)
    def on_input_submitted(self) -> None:
        self.action_refresh()

    def set_loading(self, loading: bool) -> None:
        self.query_one(DataTable).loading = loading

    def action_refresh(self) -> None:
        self.set_loading(True)

        ts_str = self.query_one("#date", Input).value
        ts = to_unix_timestamp(ts_str if ts_str else "now")

        aircraft_widget = self.query_one(AircraftWidget)
        if aircraft := aircraft_widget.aircraft_id:
            self.run_worker(
                self.lookup_aircraft(aircraft, ts=ts), exclusive=True
            )
            return
        number_widget = self.query_one(FlightWidget)
        if number := number_widget.number:
            self.run_worker(self.lookup_number(number, ts=ts), exclusive=True)
            return
        departure_widget = self.query_one("#departure", AirportWidget)
        arrival_widget = self.query_one("#arrival", AirportWidget)
        if departure := departure_widget.airport_id:
            if arrival := arrival_widget.airport_id:
                ts = get_current_timestamp() if ts == "now" else ts
                self.run_worker(
                    self.lookup_city_pair(departure, arrival, ts=ts),
                    exclusive=True,
                )
                return
            self.run_worker(
                self.lookup_departure(departure, ts=ts), exclusive=True
            )
            return
        if arrival := arrival_widget.airport_id:
            self.run_worker(self.lookup_arrival(arrival, ts=ts), exclusive=True)
            return

        self.set_loading(False)

    async def lookup_aircraft(
        self, value: str, ts: IntoTimestamp | Literal["now"]
    ) -> None:
        try:
            result = await self.fr24.flight_list.fetch(
                reg=value,
                limit=100,
                timestamp=ts,
            )
            results = result.to_dict()
            self.update_table(results["result"]["response"].get("data", None))
        except UnwrapError as e:
            self.notify(f"Error: {e}", severity="error", title="API Error")
        finally:
            self.set_loading(False)

    async def lookup_number(
        self, value: str, ts: IntoTimestamp | Literal["now"]
    ) -> None:
        try:
            result = await self.fr24.flight_list.fetch(
                flight=value,
                limit=100,
                timestamp=ts,
            )
            results = result.to_dict()
            self.update_table(results["result"]["response"].get("data", None))
        except UnwrapError as e:
            self.notify(f"Error: {e}", severity="error", title="API Error")
        finally:
            self.set_loading(False)

    async def lookup_city_pair(
        self, departure: str, arrival: str, ts: TimestampS[int]
    ) -> None:
        try:
            result = await self.fr24.find.fetch(query=f"{departure}-{arrival}")
            results = result.to_dict()
            if results is None or results["stats"]["count"]["schedule"] == 0:
                return
            flight_numbers = list(
                sched["detail"]["flight"]
                for sched in results["results"]
                if is_schedule(sched)
            )
            flight_lists: list[FlightList] = []
            for value in flight_numbers:
                try:
                    res_obj = await self.fr24.flight_list.fetch(
                        flight=value, limit=10, timestamp=ts
                    )
                    res = res_obj.to_dict()
                except UnwrapError as exc:
                    err = exc.err
                    if (
                        isinstance(err, httpx.HTTPStatusError)
                        and err.response.status_code == 402
                    ):
                        await asyncio.sleep(10)
                        res_obj = await self.fr24.flight_list.fetch(
                            flight=value, limit=10, timestamp=ts
                        )
                        res = res_obj.to_dict()
                    else:
                        raise exc

                flight_lists.append(res)

                compacted_view = list(
                    flatten(
                        *(
                            entry
                            for e in flight_lists
                            if (
                                (entry := e["result"]["response"]["data"])
                                is not None
                            )
                        )
                    )
                )

                def by_departure_time(elt: FlightListItem) -> int:
                    departure_time = elt["time"]["scheduled"]["departure"]
                    return -departure_time if departure_time else 0

                compacted_view = sorted(
                    (
                        entry
                        for entry in compacted_view
                        if (sobt := entry["time"]["scheduled"]["departure"])
                        is not None
                        and sobt < ts + 3600 * 24
                    ),
                    key=by_departure_time,
                )
                self.update_table(compacted_view)

                await asyncio.sleep(2)
        except UnwrapError as e:
            self.notify(f"Error: {e}", severity="error", title="API Error")
        finally:
            self.set_loading(False)

    async def lookup_arrival(
        self, value: str, ts: IntoTimestamp | Literal["now"]
    ) -> None:
        try:
            result = await self.fr24.airport_list.fetch(
                airport=value,
                mode="arrivals",
                limit=100,
                timestamp=ts,
            )
            results = result.to_dict()
            s = results["result"]["response"]["airport"]["pluginData"][
                "schedule"
            ]
            data = s["arrivals"].get("data", None)
            if data is not None:
                self.update_table(
                    [  # TODO add airport info from
                        elt["flight"]  # type: ignore
                        for elt in data
                    ]
                )
        except UnwrapError as e:
            self.notify(str(e), severity="error", title="API Error")
        finally:
            self.set_loading(False)

    async def lookup_departure(
        self, value: str, ts: IntoTimestamp | Literal["now"]
    ) -> None:
        try:
            result = await self.fr24.airport_list.fetch(
                airport=value,
                mode="departures",
                limit=100,
                timestamp=ts,
            )
            results = result.to_dict()
            s = results["result"]["response"]["airport"]["pluginData"][
                "schedule"
            ]
            data = s["departures"].get("data", None)
            if data is not None:
                self.update_table(
                    [  # TODO add airport info from
                        elt["flight"]  # type: ignore
                        for elt in data
                    ]
                )
        except UnwrapError as e:
            self.notify(f"Error: {e}", severity="error", title="API Error")
        finally:
            self.set_loading(False)

    def update_table(self, data: None | list[FlightListItem]) -> None:
        table = self.query_one(DataTable)
        table.clear()
        if data is None:
            return
        table.add_rows(
            [
                (
                    f"{Time(entry['time']['scheduled']['departure']):%Y-%m-%d}",
                    entry["identification"]["number"]["default"],
                    entry["identification"]["callsign"],
                    fmt_aircraft(entry.get("aircraft")),
                    fmt_airport(entry.get("airport", {}).get("origin")),
                    fmt_airport(entry.get("airport", {}).get("destination")),
                    f"{Time(entry['time']['scheduled']['departure']):%H:%MZ}",
                    f"{Time(entry['time']['real']['departure']):%H:%MZ}",
                    f"{Time(entry['time']['scheduled']['arrival']):%H:%MZ}",
                    fmt_status(entry["status"]["text"] or ""),
                    entry["identification"]["id"],
                )
                for entry in data
            ]
        )


def main() -> None:
    app = FR24Tui()
    app.run()


if __name__ == "__main__":
    main()



# ruff: noqa
# mypy: disable-error-code="top-level-await, no-redef"
# %%
# --8<-- [start:script]
from fr24 import FR24, FR24Cache


async def get_farthest_flight_id(fr24: FR24) -> int:
    nearest_result = await fr24.nearest_flights.fetch(
        lat=22.31257, lon=113.92708, radius=10000, limit=1500
    )
    return nearest_result.to_proto().flights_list[-1].flight.flightid


async def my_flight_details() -> None:
    async with FR24() as fr24:
        flight_id = await get_farthest_flight_id(fr24)
        result = await fr24.flight_details.fetch(flight_id=flight_id)
        print(result)
        print(result.to_dict())
        print(result.to_polars())
        result.write_table(FR24Cache.default())


await my_flight_details()
# --8<-- [end:script]

# %%
"""
# --8<-- [start:result]
FlightDetailsResult(
    request=FlightDetailsParams(flight_id=1011852915, restriction_mode=0, verbose=True),
    response=<Response [200 OK]>,
    timestamp=1758466410
)
# --8<-- [end:result]
# --8<-- [start:dict]
{
    "aircraft_info": {
        "icao_address": 7864687,
        "reg": "B-HNG",
        "type": "B773",
        "icon": "B777",
        "full_description": "Boeing 777-367",
        "images_list": [
            {
                "url": "https://www.jetphotos.com/photo/11821124",
                "copyright": "kuzuma",
                "thumbnail": "https://cdn.jetphotos.com/200/5/505864_1756100441_tb.jpg",
                "medium": "https://cdn.jetphotos.com/400/5/505864_1756100441.jpg",
                "large": "https://cdn.jetphotos.com/640/5/505864_1756100441.jpg",
            },
            {
                "url": "https://www.jetphotos.com/photo/11808578",
                "copyright": "tobiashsu1217_spotter",
                "thumbnail": "https://cdn.jetphotos.com/200/6/1544325_1755012630_tb.jpg",
                "medium": "https://cdn.jetphotos.com/400/6/1544325_1755012630.jpg",
                "large": "https://cdn.jetphotos.com/640/6/1544325_1755012630.jpg",
            },
            {
                "url": "https://www.jetphotos.com/photo/11804947",
                "copyright": "Stanley Joe Hidayat",
                "thumbnail": "https://cdn.jetphotos.com/200/5/389644_1754704150_tb.jpg",
                "medium": "https://cdn.jetphotos.com/400/5/389644_1754704150.jpg",
                "large": "https://cdn.jetphotos.com/640/5/389644_1754704150.jpg",
            },
        ],
        "msn_available": True,
        "age_available": True,
        "registered_owners": "Cathay Pacific",
        "is_country_of_reg_available": True,
    },
    "schedule_info": {
        "flight_number": "CX902",
        "operated_by_id": 57,
        "painted_as_id": 57,
        "origin_id": 2266,
        "destination_id": 1366,
        "scheduled_departure": 1758457200,
        "scheduled_arrival": 1758466500,
        "actual_departure": 1758460310,
        "arr_terminal": "1",
        "baggage_belt": "10",
    },
    "flight_progress": {
        "traversed_distance": 1146491,
        "remaining_distance": 3113,
        "elapsed_time": 6100,
        "remaining_time": 29,
        "eta": 1758466439,
        "great_circle_distance": 1143502,
        "mean_flight_time": 5846,
        "flight_stage": "AIRBORNE",
        "delay_status": "GREEN",
        "progress_pct": 99,
    },
    "flight_info": {
        "flightid": 1011852915,
        "lat": 22.324324,
        "lon": 113.89048,
        "track": 70,
        "speed": 120,
        "timestamp_ms": "1758466407324",
        "callsign": "CPA902",
        "ems_availability": {
            "qnh_availability": True,
            "amcp_availability": True,
            "oat_availability": True,
            "ias_availability": True,
            "tas_availability": True,
            "mach_availability": True,
            "agps_availability": True,
            "agpsdiff_availability": True,
            "wind_dir_availability": True,
            "wind_speed_availability": True,
        },
        "squawk_availability": True,
        "airspace_availability": True,
        "server_time_ms": "1758466410387",
    },
    "flight_plan": {},
    "flight_trail_list": [
        {
            "snapshot_id": "1758458895",
            "lat": 14.517563,
            "lon": 121.01436,
            "spd": 1,
            "heading": 101,
        },
        ...
        {
            "snapshot_id": "1758466399",
            "lat": 22.322754,
            "lon": 113.88553,
            "altitude": 100,
            "spd": 141,
            "heading": 70,
        },
    ],
    "position_buffer": {
        "recent_positions_list": [
            {"delta_lat": 43, "delta_lon": 131, "delta_ms": 2079},
            {"delta_lat": 94, "delta_lon": 294, "delta_ms": 4660},
            {"delta_lat": 112, "delta_lon": 350, "delta_ms": 5589},
            {"delta_lat": 156, "delta_lon": 495, "delta_ms": 7660},
            {"delta_lat": 196, "delta_lon": 614, "delta_ms": 9579},
            {"delta_lat": 234, "delta_lon": 721, "delta_ms": 11019},
        ]
    },
}
# --8<-- [end:dict]
# --8<-- [start:polars]
shape: (1, 32)
┌─────────────┬───────┬──────────┬─────────────┬───┬────────┬────────────┬────────────┬────────────┐
│ icao_addres ┆ reg   ┆ typecode ┆ flight_numb ┆ … ┆ squawk ┆ ems        ┆ flight_tra ┆ position_b │
│ s           ┆ ---   ┆ ---      ┆ er          ┆   ┆ ---    ┆ ---        ┆ il_list    ┆ uffer      │
│ ---         ┆ str   ┆ str      ┆ ---         ┆   ┆ u16    ┆ struct[13] ┆ ---        ┆ ---        │
│ u32         ┆       ┆          ┆ str         ┆   ┆        ┆            ┆ list[struc ┆ list[struc │
│             ┆       ┆          ┆             ┆   ┆        ┆            ┆ t[8]]      ┆ t[3]]      │
╞═════════════╪═══════╪══════════╪═════════════╪═══╪════════╪════════════╪════════════╪════════════╡
│ 7864687     ┆ B-HNG ┆ B773     ┆ CX902       ┆ … ┆ 0      ┆ {0,0,0,0,0 ┆ [{17584588 ┆ [{43,131,2 │
│             ┆       ┆          ┆             ┆   ┆        ┆ ,0,0,0,0,0 ┆ 95,14.5175 ┆ 079}, {94, │
│             ┆       ┆          ┆             ┆   ┆        ┆ ,0,0,0}    ┆ 63,121.014 ┆ 294,4660}, │
│             ┆       ┆          ┆             ┆   ┆        ┆            ┆ …          ┆ …          │
└─────────────┴───────┴──────────┴─────────────┴───┴────────┴────────────┴────────────┴────────────┘
# --8<-- [end:polars]
"""



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: fr24/proto/_common.proto
# Protobuf Python Version: 6.33.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    0,
    '',
    'fr24/proto/_common.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18\x66r24/proto/_common.proto\x12\x07_common\"\xca\x01\n\x07\x45MSInfo\x12\x0b\n\x03qnh\x18\x01 \x01(\x05\x12\x0c\n\x04\x61mcp\x18\x02 \x01(\x05\x12\x0c\n\x04\x61\x66ms\x18\x03 \x01(\x05\x12\x0b\n\x03oat\x18\x04 \x01(\x05\x12\x0b\n\x03ias\x18\x05 \x01(\x05\x12\x0b\n\x03tas\x18\x06 \x01(\x05\x12\x0c\n\x04mach\x18\x07 \x01(\x05\x12\x0c\n\x04\x61gps\x18\x08 \x01(\x05\x12\x10\n\x08\x61gpsdiff\x18\t \x01(\x05\x12\x0f\n\x07\x61pflags\x18\n \x01(\x05\x12\x10\n\x08wind_dir\x18\x0b \x01(\x05\x12\x12\n\nwind_speed\x18\x0c \x01(\x05\x12\n\n\x02rs\x18\r \x01(\x05\"\xfb\x02\n\x0f\x45MSAvailability\x12\x18\n\x10qnh_availability\x18\x01 \x01(\x08\x12\x19\n\x11\x61mcp_availability\x18\x02 \x01(\x08\x12\x19\n\x11\x61\x66ms_availability\x18\x03 \x01(\x08\x12\x18\n\x10oat_availability\x18\x04 \x01(\x08\x12\x18\n\x10ias_availability\x18\x05 \x01(\x08\x12\x18\n\x10tas_availability\x18\x06 \x01(\x08\x12\x19\n\x11mach_availability\x18\x07 \x01(\x08\x12\x19\n\x11\x61gps_availability\x18\x08 \x01(\x08\x12\x1d\n\x15\x61gpsdiff_availability\x18\t \x01(\x08\x12\x1c\n\x14\x61pflags_availability\x18\n \x01(\x08\x12\x1d\n\x15wind_dir_availability\x18\x0b \x01(\x08\x12\x1f\n\x17wind_speed_availability\x18\x0c \x01(\x08\x12\x17\n\x0frs_availability\x18\r \x01(\x08\"n\n\x08Schedule\x12\x0b\n\x03std\x18\x01 \x01(\x05\x12\x0b\n\x03\x65td\x18\x02 \x01(\x05\x12\x0b\n\x03\x61td\x18\x03 \x01(\x05\x12\x0b\n\x03sta\x18\x04 \x01(\x05\x12\x0b\n\x03\x65ta\x18\x05 \x01(\x05\x12\x0b\n\x03\x61ta\x18\x06 \x01(\x05\x12\x14\n\x0cprogress_pct\x18\x07 \x01(\x05\"6\n\x05Route\x12\x0c\n\x04\x66rom\x18\x01 \x01(\t\x12\n\n\x02to\x18\x02 \x01(\t\x12\x13\n\x0b\x64iverted_to\x18\x03 \x01(\t\"\xdc\x03\n\x0f\x45xtraFlightInfo\x12\x0e\n\x06\x66light\x18\x01 \x01(\t\x12\x0b\n\x03reg\x18\x02 \x01(\t\x12\x1d\n\x05route\x18\x03 \x01(\x0b\x32\x0e._common.Route\x12\x0c\n\x04type\x18\x04 \x01(\t\x12\x0e\n\x06squawk\x18\x05 \x01(\x05\x12\x0e\n\x06vspeed\x18\x06 \x01(\x05\x12\x0b\n\x03\x61ge\x18\x07 \x01(\t\x12\x16\n\x0e\x63ountry_of_reg\x18\x08 \x01(\x05\x12#\n\x08schedule\x18\t \x01(\x0b\x32\x11._common.Schedule\x12\x0f\n\x07logo_id\x18\n \x01(\x05\x12\x10\n\x08\x61irspace\x18\x0b \x01(\x05\x12\"\n\x08\x65ms_info\x18\x0c \x01(\x0b\x32\x10._common.EMSInfo\x12\x32\n\x10\x65ms_availability\x18\r \x01(\x0b\x32\x18._common.EMSAvailability\x12\x14\n\x0cicao_address\x18\x0e \x01(\r\x12\x16\n\x0eoperated_by_id\x18\x0f \x01(\r\x12\x1b\n\x13squawk_availability\x18\x10 \x01(\x08\x12\x1b\n\x13vspeed_availability\x18\x11 \x01(\x08\x12\x1d\n\x15\x61irspace_availability\x18\x12 \x01(\x08\x12\x13\n\x0b\x61irspace_id\x18\x13 \x01(\t\"A\n\x0bSourceStats\x12#\n\x06source\x18\x01 \x01(\x0e\x32\x13._common.DataSource\x12\r\n\x05\x63ount\x18\x02 \x01(\r\"1\n\x05Stats\x12(\n\ntotal_list\x18\x01 \x03(\x0b\x32\x14._common.SourceStats\"H\n\x0eRecentPosition\x12\x11\n\tdelta_lat\x18\x01 \x01(\x05\x12\x11\n\tdelta_lon\x18\x02 \x01(\x05\x12\x10\n\x08\x64\x65lta_ms\x18\x03 \x01(\r\"H\n\x0ePositionBuffer\x12\x36\n\x15recent_positions_list\x18\x01 \x03(\x0b\x32\x17._common.RecentPosition\"\xf0\x02\n\x06\x46light\x12\x10\n\x08\x66lightid\x18\x01 \x01(\x05\x12\x0b\n\x03lat\x18\x02 \x01(\x02\x12\x0b\n\x03lon\x18\x03 \x01(\x02\x12\r\n\x05track\x18\x04 \x01(\x05\x12\x0b\n\x03\x61lt\x18\x05 \x01(\x05\x12\r\n\x05speed\x18\x06 \x01(\x05\x12\x1b\n\x04icon\x18\x07 \x01(\x0e\x32\r._common.Icon\x12\x1f\n\x06status\x18\x08 \x01(\x0e\x32\x0f._common.Status\x12\x11\n\ttimestamp\x18\t \x01(\x05\x12\x11\n\ton_ground\x18\n \x01(\x08\x12\x10\n\x08\x63\x61llsign\x18\x0b \x01(\t\x12#\n\x06source\x18\x0c \x01(\x0e\x32\x13._common.DataSource\x12,\n\nextra_info\x18\r \x01(\x0b\x32\x18._common.ExtraFlightInfo\x12\x30\n\x0fposition_buffer\x18\x0e \x01(\x0b\x32\x17._common.PositionBuffer\x12\x14\n\x0ctimestamp_ms\x18\x0f \x01(\x04\"\xaa\x02\n\x0e\x46lightProgress\x12\x1a\n\x12traversed_distance\x18\x01 \x01(\r\x12\x1a\n\x12remaining_distance\x18\x02 \x01(\r\x12\x14\n\x0c\x65lapsed_time\x18\x03 \x01(\r\x12\x16\n\x0eremaining_time\x18\x04 \x01(\r\x12\x0b\n\x03\x65ta\x18\x05 \x01(\r\x12\x1d\n\x15great_circle_distance\x18\x06 \x01(\r\x12\x18\n\x10mean_flight_time\x18\x07 \x01(\r\x12*\n\x0c\x66light_stage\x18\x08 \x01(\x0e\x32\x14._common.FlightStage\x12*\n\x0c\x64\x65lay_status\x18\t \x01(\x0e\x32\x14._common.DelayStatus\x12\x14\n\x0cprogress_pct\x18\n \x01(\r\"o\n\tImageInfo\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x11\n\tcopyright\x18\x02 \x01(\t\x12\x11\n\tthumbnail\x18\x03 \x01(\t\x12\x0e\n\x06medium\x18\x04 \x01(\t\x12\r\n\x05large\x18\x05 \x01(\t\x12\x10\n\x08sideview\x18\x06 \x01(\t\"\x94\x03\n\x0c\x41ircraftInfo\x12\x14\n\x0cicao_address\x18\x01 \x01(\r\x12\x0b\n\x03reg\x18\x02 \x01(\t\x12\x16\n\x0e\x63ountry_of_reg\x18\x03 \x01(\x05\x12\x0c\n\x04type\x18\x04 \x01(\t\x12\x1b\n\x04icon\x18\x05 \x01(\x0e\x32\r._common.Icon\x12\x18\n\x10\x66ull_description\x18\x06 \x01(\t\x12\x0b\n\x03msn\x18\x07 \x01(\t\x12!\n\x07service\x18\x08 \x01(\x0e\x32\x10._common.Service\x12\x15\n\rac_birth_date\x18\t \x01(\t\x12\x0e\n\x06\x61\x63_age\x18\n \x01(\r\x12\'\n\x0bimages_list\x18\x0b \x03(\x0b\x32\x12._common.ImageInfo\x12\x16\n\x0eis_test_flight\x18\x0c \x01(\x08\x12\x15\n\rmsn_available\x18\r \x01(\x08\x12\x15\n\rage_available\x18\x0e \x01(\x08\x12\x19\n\x11registered_owners\x18\x0f \x01(\t\x12#\n\x1bis_country_of_reg_available\x18\x10 \x01(\x08\",\n\x05Point\x12\x10\n\x08latitude\x18\x01 \x01(\x05\x12\x11\n\tlongitude\x18\x02 \x01(\x05\"9\n\nCoordinate\x12\x0c\n\x04\x63ode\x18\x01 \x01(\t\x12\x1d\n\x05point\x18\x02 \x01(\x0b\x32\x0e._common.Point\"M\n\x03\x46ix\x12\x0f\n\x07\x61irport\x18\x01 \x01(\t\x12\x0c\n\x04\x61rea\x18\x02 \x01(\t\x12\'\n\ncoordinate\x18\x03 \x01(\x0b\x32\x13._common.Coordinate\";\n\nAltArrival\x12\x1d\n\x07\x61rrival\x18\x01 \x01(\x0b\x32\x0c._common.Fix\x12\x0e\n\x06length\x18\x02 \x01(\x02\"\xde\x01\n\nFlightPlan\x12\x11\n\tdeparture\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65stination\x18\x02 \x01(\t\x12\x18\n\x10\x66light_plan_icao\x18\x03 \x01(\t\x12\x0e\n\x06length\x18\x04 \x01(\x01\x12*\n\ralt_arrival_1\x18\x05 \x01(\x0b\x32\x13._common.AltArrival\x12*\n\ralt_arrival_2\x18\x06 \x01(\x0b\x32\x13._common.AltArrival\x12&\n\x0ewaypoints_list\x18\x07 \x03(\x0b\x32\x0e._common.Point\"\xbf\x02\n\x0cScheduleInfo\x12\x15\n\rflight_number\x18\x01 \x01(\t\x12\x16\n\x0eoperated_by_id\x18\x02 \x01(\r\x12\x15\n\rpainted_as_id\x18\x03 \x01(\r\x12\x11\n\torigin_id\x18\x04 \x01(\r\x12\x16\n\x0e\x64\x65stination_id\x18\x05 \x01(\r\x12\x16\n\x0e\x64iverted_to_id\x18\x06 \x01(\r\x12\x1b\n\x13scheduled_departure\x18\x07 \x01(\r\x12\x19\n\x11scheduled_arrival\x18\x08 \x01(\r\x12\x18\n\x10\x61\x63tual_departure\x18\t \x01(\r\x12\x16\n\x0e\x61\x63tual_arrival\x18\n \x01(\r\x12\x14\n\x0c\x61rr_terminal\x18\x0b \x01(\t\x12\x10\n\x08\x61rr_gate\x18\x0c \x01(\t\x12\x14\n\x0c\x62\x61ggage_belt\x18\r \x01(\t\"\xfc\x03\n\x12\x45xtendedFlightInfo\x12\x10\n\x08\x66lightid\x18\x01 \x01(\r\x12\x0b\n\x03lat\x18\x02 \x01(\x02\x12\x0b\n\x03lon\x18\x03 \x01(\x02\x12\r\n\x05track\x18\x04 \x01(\x05\x12\x0b\n\x03\x61lt\x18\x05 \x01(\x05\x12\r\n\x05speed\x18\x06 \x01(\x05\x12\x1f\n\x06status\x18\x07 \x01(\x0e\x32\x0f._common.Status\x12\x14\n\x0ctimestamp_ms\x18\x08 \x01(\x04\x12\x11\n\ton_ground\x18\t \x01(\x08\x12\x10\n\x08\x63\x61llsign\x18\n \x01(\t\x12#\n\x06source\x18\x0b \x01(\x0e\x32\x13._common.DataSource\x12\x32\n\x10\x65ms_availability\x18\x0c \x01(\x0b\x32\x18._common.EMSAvailability\x12\"\n\x08\x65ms_info\x18\r \x01(\x0b\x32\x10._common.EMSInfo\x12\x1b\n\x13squawk_availability\x18\x0e \x01(\x08\x12\x0e\n\x06squawk\x18\x0f \x01(\x05\x12\x1b\n\x13vspeed_availability\x18\x10 \x01(\x08\x12\x0e\n\x06vspeed\x18\x11 \x01(\x05\x12\x1d\n\x15\x61irspace_availability\x18\x12 \x01(\x08\x12\x10\n\x08\x61irspace\x18\x13 \x01(\t\x12\x13\n\x0b\x61irspace_id\x18\x14 \x01(\t\x12\x16\n\x0eserver_time_ms\x18\x15 \x01(\x04\"\x9e\x01\n\nTrailPoint\x12\x13\n\x0bsnapshot_id\x18\x01 \x01(\x04\x12\x0b\n\x03lat\x18\x02 \x01(\x02\x12\x0b\n\x03lon\x18\x03 \x01(\x02\x12\x10\n\x08\x61ltitude\x18\x04 \x01(\x05\x12\x0b\n\x03spd\x18\x05 \x01(\r\x12\x0f\n\x07heading\x18\x06 \x01(\r\x12\x0c\n\x04vspd\x18\x07 \x01(\x05\x12#\n\x06source\x18\x08 \x01(\x0e\x32\x13._common.DataSource\"\x19\n\x08\x44uration\x12\r\n\x05\x63ount\x18\x01 \x01(\r\"\x15\n\x04Tick\x12\r\n\x05\x63ount\x18\x01 \x01(\r\"\xc6\x01\n\x12RadarHistoryRecord\x12\x11\n\ttimestamp\x18\x01 \x01(\x04\x12\x0b\n\x03lat\x18\x02 \x01(\x02\x12\x0b\n\x03lon\x18\x03 \x01(\x02\x12\x10\n\x08\x61ltitude\x18\x04 \x01(\x05\x12\x0b\n\x03spd\x18\x05 \x01(\r\x12\x0f\n\x07heading\x18\x06 \x01(\r\x12\x0c\n\x04vspd\x18\x07 \x01(\x05\x12\x0e\n\x06squawk\x18\x08 \x01(\r\x12#\n\x06source\x18\t \x01(\x0e\x32\x13._common.DataSource\x12\x10\n\x08\x63\x61llsign\x18\n \x01(\t*R\n\x15RestrictionVisibility\x12\x0f\n\x0bNOT_VISIBLE\x10\x00\x12\x15\n\x11PARTIALLY_VISIBLE\x10\x01\x12\x11\n\rFULLY_VISIBLE\x10\x02*\xe6\x01\n\x07Service\x12\r\n\tPASSENGER\x10\x00\x12\t\n\x05\x43\x41RGO\x10\x01\x12\x1b\n\x17MILITARY_AND_GOVERNMENT\x10\x02\x12\x11\n\rBUSINESS_JETS\x10\x03\x12\x14\n\x10GENERAL_AVIATION\x10\x04\x12\x0f\n\x0bHELICOPTERS\x10\x05\x12\x14\n\x10LIGHTER_THAN_AIR\x10\x06\x12\x0b\n\x07GLIDERS\x10\x07\x12\n\n\x06\x44RONES\x10\x08\x12\x13\n\x0fGROUND_VEHICLES\x10\t\x12\x11\n\rOTHER_SERVICE\x10\n\x12\x13\n\x0fNON_CATEGORIZED\x10\x0b*D\n\x0bTrafficType\x12\x08\n\x04NONE\x10\x00\x12\x0f\n\x0bGROUND_ONLY\x10\x01\x12\x11\n\rAIRBORNE_ONLY\x10\x02\x12\x07\n\x03\x41LL\x10\x03*\x99\x01\n\nDataSource\x12\x08\n\x04\x41\x44SB\x10\x00\x12\x08\n\x04MLAT\x10\x01\x12\t\n\x05\x46LARM\x10\x02\x12\x07\n\x03\x46\x41\x41\x10\x03\x12\r\n\tESTIMATED\x10\x04\x12\r\n\tSATELLITE\x10\x05\x12\x15\n\x11OTHER_DATA_SOURCE\x10\x06\x12\x07\n\x03UAT\x10\x07\x12\x10\n\x0cSPIDERTRACKS\x10\x08\x12\x07\n\x03\x41US\x10\t\x12\n\n\x06\x41IERON\x10\n*\xbd\x02\n\x04Icon\x12\x08\n\x04\x42\x37\x33\x38\x10\x00\x12\x08\n\x04\x46GTR\x10\x01\x12\t\n\x05\x41SW20\x10\x02\x12\x08\n\x04\x43\x32\x30\x36\x10\x03\x12\x08\n\x04\x43\x33\x30\x33\x10\x04\x12\x08\n\x04LJ60\x10\x05\x12\x08\n\x04Q300\x10\x06\x12\x08\n\x04\x42\x37\x33\x36\x10\x07\x12\r\n\tFOKKER100\x10\x08\x12\x08\n\x04RJ85\x10\t\x12\x08\n\x04\x41\x33\x32\x30\x10\n\x12\x08\n\x04\x42\x37\x35\x37\x10\x0b\x12\x08\n\x04\x42\x37\x36\x37\x10\x0c\x12\x08\n\x04\x41\x33ST\x10\r\x12\x08\n\x04MD11\x10\x0e\x12\x08\n\x04\x41\x33\x33\x30\x10\x0f\x12\x08\n\x04\x41\x33\x34\x33\x10\x10\x12\x08\n\x04\x41\x33\x34\x36\x10\x11\x12\x08\n\x04\x42\x37\x37\x37\x10\x12\x12\x08\n\x04\x42\x37\x34\x37\x10\x13\x12\x08\n\x04\x41\x33\x38\x30\x10\x14\x12\x08\n\x04\x41\x32\x32\x35\x10\x15\x12\x07\n\x03SI2\x10\x16\x12\x06\n\x02\x45\x43\x10\x17\x12\x08\n\x04\x42\x41LL\x10\x18\x12\x08\n\x04GRND\x10\x19\x12\x08\n\x04SLEI\x10\x1a\x12\x08\n\x04\x44RON\x10\x1b\x12\x07\n\x03SAT\x10\x1c\x12\x07\n\x03ISS\x10\x1d\x12\x08\n\x04\x43ONC\x10\x1e*P\n\x06Status\x12\n\n\x06NORMAL\x10\x00\x12\x0e\n\nBACKGROUND\x10\x01\x12\r\n\tEMERGENCY\x10\x02\x12\x11\n\rNOT_AVAILABLE\x10\x03\x12\x08\n\x04LIVE\x10\x04*e\n\x0b\x46lightStage\x12\x0b\n\x07UNKNOWN\x10\x00\x12\r\n\tON_GROUND\x10\x01\x12\r\n\tASCENDING\x10\x02\x12\x0c\n\x08\x41IRBORNE\x10\x03\x12\x0e\n\nDESCENDING\x10\x04\x12\r\n\tDIVERSION\x10\x05*7\n\x0b\x44\x65layStatus\x12\x08\n\x04GRAY\x10\x00\x12\t\n\x05GREEN\x10\x01\x12\n\n\x06YELLOW\x10\x02\x12\x07\n\x03RED\x10\x03*\xc2\x01\n\x0f\x45mergencyStatus\x12\x10\n\x0cNO_EMERGENCY\x10\x00\x12\x15\n\x11GENERAL_EMERGENCY\x10\x01\x12\x1f\n\x1bLIFEGUARD_MEDICAL_EMERGENCY\x10\x02\x12\x10\n\x0cMINIMUM_FUEL\x10\x03\x12\x15\n\x11NO_COMMUNICATIONS\x10\x04\x12\x19\n\x15UNLAWFUL_INTERFERENCE\x10\x05\x12\x13\n\x0f\x44OWNED_AIRCRAFT\x10\x06\x12\x0c\n\x08RESERVED\x10\x07\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'fr24.proto._common_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_RESTRICTIONVISIBILITY']._serialized_start=4444
  _globals['_RESTRICTIONVISIBILITY']._serialized_end=4526
  _globals['_SERVICE']._serialized_start=4529
  _globals['_SERVICE']._serialized_end=4759
  _globals['_TRAFFICTYPE']._serialized_start=4761
  _globals['_TRAFFICTYPE']._serialized_end=4829
  _globals['_DATASOURCE']._serialized_start=4832
  _globals['_DATASOURCE']._serialized_end=4985
  _globals['_ICON']._serialized_start=4988
  _globals['_ICON']._serialized_end=5305
  _globals['_STATUS']._serialized_start=5307
  _globals['_STATUS']._serialized_end=5387
  _globals['_FLIGHTSTAGE']._serialized_start=5389
  _globals['_FLIGHTSTAGE']._serialized_end=5490
  _globals['_DELAYSTATUS']._serialized_start=5492
  _globals['_DELAYSTATUS']._serialized_end=5547
  _globals['_EMERGENCYSTATUS']._serialized_start=5550
  _globals['_EMERGENCYSTATUS']._serialized_end=5744
  _globals['_EMSINFO']._serialized_start=38
  _globals['_EMSINFO']._serialized_end=240
  _globals['_EMSAVAILABILITY']._serialized_start=243
  _globals['_EMSAVAILABILITY']._serialized_end=622
  _globals['_SCHEDULE']._serialized_start=624
  _globals['_SCHEDULE']._serialized_end=734
  _globals['_ROUTE']._serialized_start=736
  _globals['_ROUTE']._serialized_end=790
  _globals['_EXTRAFLIGHTINFO']._serialized_start=793
  _globals['_EXTRAFLIGHTINFO']._serialized_end=1269
  _globals['_SOURCESTATS']._serialized_start=1271
  _globals['_SOURCESTATS']._serialized_end=1336
  _globals['_STATS']._serialized_start=1338
  _globals['_STATS']._serialized_end=1387
  _globals['_RECENTPOSITION']._serialized_start=1389
  _globals['_RECENTPOSITION']._serialized_end=1461
  _globals['_POSITIONBUFFER']._serialized_start=1463
  _globals['_POSITIONBUFFER']._serialized_end=1535
  _globals['_FLIGHT']._serialized_start=1538
  _globals['_FLIGHT']._serialized_end=1906
  _globals['_FLIGHTPROGRESS']._serialized_start=1909
  _globals['_FLIGHTPROGRESS']._serialized_end=2207
  _globals['_IMAGEINFO']._serialized_start=2209
  _globals['_IMAGEINFO']._serialized_end=2320
  _globals['_AIRCRAFTINFO']._serialized_start=2323
  _globals['_AIRCRAFTINFO']._serialized_end=2727
  _globals['_POINT']._serialized_start=2729
  _globals['_POINT']._serialized_end=2773
  _globals['_COORDINATE']._serialized_start=2775
  _globals['_COORDINATE']._serialized_end=2832
  _globals['_FIX']._serialized_start=2834
  _globals['_FIX']._serialized_end=2911
  _globals['_ALTARRIVAL']._serialized_start=2913
  _globals['_ALTARRIVAL']._serialized_end=2972
  _globals['_FLIGHTPLAN']._serialized_start=2975
  _globals['_FLIGHTPLAN']._serialized_end=3197
  _globals['_SCHEDULEINFO']._serialized_start=3200
  _globals['_SCHEDULEINFO']._serialized_end=3519
  _globals['_EXTENDEDFLIGHTINFO']._serialized_start=3522
  _globals['_EXTENDEDFLIGHTINFO']._serialized_end=4030
  _globals['_TRAILPOINT']._serialized_start=4033
  _globals['_TRAILPOINT']._serialized_end=4191
  _globals['_DURATION']._serialized_start=4193
  _globals['_DURATION']._serialized_end=4218
  _globals['_TICK']._serialized_start=4220
  _globals['_TICK']._serialized_end=4241
  _globals['_RADARHISTORYRECORD']._serialized_start=4244
  _globals['_RADARHISTORYRECORD']._serialized_end=4442
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: fr24/proto/_flight_details.proto
# Protobuf Python Version: 6.33.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    0,
    '',
    'fr24/proto/_flight_details.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from fr24.proto import _common_pb2 as fr24_dot_proto_dot___common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n fr24/proto/_flight_details.proto\x1a\x18\x66r24/proto/_common.proto\"\xaa\x02\n\x0e\x46lightProgress\x12\x1a\n\x12traversed_distance\x18\x01 \x01(\r\x12\x1a\n\x12remaining_distance\x18\x02 \x01(\r\x12\x14\n\x0c\x65lapsed_time\x18\x03 \x01(\x05\x12\x16\n\x0eremaining_time\x18\x04 \x01(\x05\x12\x0b\n\x03\x65ta\x18\x05 \x01(\r\x12\x1d\n\x15great_circle_distance\x18\x06 \x01(\r\x12\x18\n\x10mean_flight_time\x18\x07 \x01(\r\x12*\n\x0c\x66light_stage\x18\x08 \x01(\x0e\x32\x14._common.FlightStage\x12*\n\x0c\x64\x65lay_status\x18\t \x01(\x0e\x32\x14._common.DelayStatus\x12\x14\n\x0cprogress_pct\x18\n \x01(\r\"t\n\x14\x46lightDetailsRequest\x12\x11\n\tflight_id\x18\x01 \x01(\x07\x12\x38\n\x10restriction_mode\x18\x02 \x01(\x0e\x32\x1e._common.RestrictionVisibility\x12\x0f\n\x07verbose\x18\x03 \x01(\x08\"\xe3\x02\n\x15\x46lightDetailsResponse\x12,\n\raircraft_info\x18\x01 \x01(\x0b\x32\x15._common.AircraftInfo\x12,\n\rschedule_info\x18\x02 \x01(\x0b\x32\x15._common.ScheduleInfo\x12\x30\n\x0f\x66light_progress\x18\x03 \x01(\x0b\x32\x17._common.FlightProgress\x12\x30\n\x0b\x66light_info\x18\x04 \x01(\x0b\x32\x1b._common.ExtendedFlightInfo\x12(\n\x0b\x66light_plan\x18\x05 \x01(\x0b\x32\x13._common.FlightPlan\x12.\n\x11\x66light_trail_list\x18\x06 \x03(\x0b\x32\x13._common.TrailPoint\x12\x30\n\x0fposition_buffer\x18\x07 \x01(\x0b\x32\x17._common.PositionBufferb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'fr24.proto._flight_details_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_FLIGHTPROGRESS']._serialized_start=63
  _globals['_FLIGHTPROGRESS']._serialized_end=361
  _globals['_FLIGHTDETAILSREQUEST']._serialized_start=363
  _globals['_FLIGHTDETAILSREQUEST']._serialized_end=479
  _globals['_FLIGHTDETAILSRESPONSE']._serialized_start=482
  _globals['_FLIGHTDETAILSRESPONSE']._serialized_end=837
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from collections import abc as abc
from fr24.proto import _common_pb2 as _common_pb2
from google.protobuf import descriptor as descriptor
from google.protobuf import message as message
from google.protobuf.internal import containers as containers
import builtins as builtins
import sys
import typing as typing

if sys.version_info >= (3, 10):
    from typing import TypeAlias as _TypeAlias
else:
    from typing_extensions import TypeAlias as _TypeAlias

DESCRIPTOR: descriptor.FileDescriptor

@typing.final
class PlaybackFlightRequest(message.Message):
    DESCRIPTOR: descriptor.Descriptor

    FLIGHT_ID_FIELD_NUMBER: builtins.int
    TIMESTAMP_FIELD_NUMBER: builtins.int
    RESTRICTION_MODE_FIELD_NUMBER: builtins.int
    flight_id: builtins.int
    """FR24 Flight ID (e.g. `962788562` = `0x3962fcd2`)"""
    timestamp: builtins.int
    """Actual time of departure (ATD) of the historic flight,
    Unix timestamp in seconds.
    """
    restriction_mode: _common_pb2.RestrictionVisibility.ValueType
    def __init__(
        self,
        *,
        flight_id: builtins.int = ...,
        timestamp: builtins.int = ...,
        restriction_mode: _common_pb2.RestrictionVisibility.ValueType = ...,
    ) -> None: ...
    _ClearFieldArgType: _TypeAlias = typing.Literal["flight_id", b"flight_id", "restriction_mode", b"restriction_mode", "timestamp", b"timestamp"]  # noqa: Y015
    def ClearField(self, field_name: _ClearFieldArgType) -> None: ...

@typing.final
class PlaybackFlightResponse(message.Message):
    DESCRIPTOR: descriptor.Descriptor

    AIRCRAFT_INFO_FIELD_NUMBER: builtins.int
    SCHEDULE_INFO_FIELD_NUMBER: builtins.int
    FLIGHT_INFO_FIELD_NUMBER: builtins.int
    FLIGHT_TRAIL_LIST_FIELD_NUMBER: builtins.int
    FLIGHT_PROGRESS_FIELD_NUMBER: builtins.int
    @builtins.property
    def aircraft_info(self) -> _common_pb2.AircraftInfo: ...
    @builtins.property
    def schedule_info(self) -> _common_pb2.ScheduleInfo: ...
    @builtins.property
    def flight_info(self) -> _common_pb2.ExtendedFlightInfo: ...
    @builtins.property
    def flight_trail_list(self) -> containers.RepeatedCompositeFieldContainer[_common_pb2.TrailPoint]: ...
    @builtins.property
    def flight_progress(self) -> _common_pb2.FlightProgress: ...
    def __init__(
        self,
        *,
        aircraft_info: _common_pb2.AircraftInfo | None = ...,
        schedule_info: _common_pb2.ScheduleInfo | None = ...,
        flight_info: _common_pb2.ExtendedFlightInfo | None = ...,
        flight_trail_list: abc.Iterable[_common_pb2.TrailPoint] | None = ...,
        flight_progress: _common_pb2.FlightProgress | None = ...,
    ) -> None: ...
    _HasFieldArgType: _TypeAlias = typing.Literal["aircraft_info", b"aircraft_info", "flight_info", b"flight_info", "flight_progress", b"flight_progress", "schedule_info", b"schedule_info"]  # noqa: Y015
    def HasField(self, field_name: _HasFieldArgType) -> builtins.bool: ...
    _ClearFieldArgType: _TypeAlias = typing.Literal["aircraft_info", b"aircraft_info", "flight_info", b"flight_info", "flight_progress", b"flight_progress", "flight_trail_list", b"flight_trail_list", "schedule_info", b"schedule_info"]  # noqa: Y015
    def ClearField(self, field_name: _ClearFieldArgType) -> None: ...



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from fr24.proto._common_pb2 import A225 as A225
from fr24.proto._common_pb2 import A320 as A320
from fr24.proto._common_pb2 import A330 as A330
from fr24.proto._common_pb2 import A343 as A343
from fr24.proto._common_pb2 import A346 as A346
from fr24.proto._common_pb2 import A380 as A380
from fr24.proto._common_pb2 import A3ST as A3ST
from fr24.proto._common_pb2 import ADSB as ADSB
from fr24.proto._common_pb2 import AIERON as AIERON
from fr24.proto._common_pb2 import AIRBORNE as AIRBORNE
from fr24.proto._common_pb2 import AIRBORNE_ONLY as AIRBORNE_ONLY
from fr24.proto._common_pb2 import ALL as ALL
from fr24.proto._common_pb2 import ASCENDING as ASCENDING
from fr24.proto._common_pb2 import ASW20 as ASW20
from fr24.proto._common_pb2 import AUS as AUS
from fr24.proto._common_pb2 import AircraftInfo as AircraftInfo
from fr24.proto._common_pb2 import AltArrival as AltArrival
from fr24.proto._common_pb2 import B736 as B736
from fr24.proto._common_pb2 import B738 as B738
from fr24.proto._common_pb2 import B747 as B747
from fr24.proto._common_pb2 import B757 as B757
from fr24.proto._common_pb2 import B767 as B767
from fr24.proto._common_pb2 import B777 as B777
from fr24.proto._common_pb2 import BACKGROUND as BACKGROUND
from fr24.proto._common_pb2 import BALL as BALL
from fr24.proto._common_pb2 import BUSINESS_JETS as BUSINESS_JETS
from fr24.proto._common_pb2 import C206 as C206
from fr24.proto._common_pb2 import C303 as C303
from fr24.proto._common_pb2 import CARGO as CARGO
from fr24.proto._common_pb2 import CONC as CONC
from fr24.proto._common_pb2 import Coordinate as Coordinate
from fr24.proto._common_pb2 import DESCENDING as DESCENDING
from fr24.proto._common_pb2 import DIVERSION as DIVERSION
from fr24.proto._common_pb2 import DOWNED_AIRCRAFT as DOWNED_AIRCRAFT
from fr24.proto._common_pb2 import DRON as DRON
from fr24.proto._common_pb2 import DRONES as DRONES
from fr24.proto._common_pb2 import DataSource as DataSource
from fr24.proto._common_pb2 import DelayStatus as DelayStatus
from fr24.proto._common_pb2 import Duration as Duration
from fr24.proto._common_pb2 import EC as EC
from fr24.proto._common_pb2 import EMERGENCY as EMERGENCY
from fr24.proto._common_pb2 import EMSAvailability as EMSAvailability
from fr24.proto._common_pb2 import EMSInfo as EMSInfo
from fr24.proto._common_pb2 import ESTIMATED as ESTIMATED
from fr24.proto._common_pb2 import EmergencyStatus as EmergencyStatus
from fr24.proto._common_pb2 import ExtendedFlightInfo as ExtendedFlightInfo
from fr24.proto._common_pb2 import ExtraFlightInfo as ExtraFlightInfo
from fr24.proto._common_pb2 import FAA as FAA
from fr24.proto._common_pb2 import FGTR as FGTR
from fr24.proto._common_pb2 import FLARM as FLARM
from fr24.proto._common_pb2 import FOKKER100 as FOKKER100
from fr24.proto._common_pb2 import FULLY_VISIBLE as FULLY_VISIBLE
from fr24.proto._common_pb2 import Fix as Fix
from fr24.proto._common_pb2 import Flight as Flight
from fr24.proto._common_pb2 import FlightPlan as FlightPlan
from fr24.proto._common_pb2 import FlightProgress as FlightProgress
from fr24.proto._common_pb2 import FlightStage as FlightStage
from fr24.proto._common_pb2 import GENERAL_AVIATION as GENERAL_AVIATION
from fr24.proto._common_pb2 import GENERAL_EMERGENCY as GENERAL_EMERGENCY
from fr24.proto._common_pb2 import GLIDERS as GLIDERS
from fr24.proto._common_pb2 import GRAY as GRAY
from fr24.proto._common_pb2 import GREEN as GREEN
from fr24.proto._common_pb2 import GRND as GRND
from fr24.proto._common_pb2 import GROUND_ONLY as GROUND_ONLY
from fr24.proto._common_pb2 import GROUND_VEHICLES as GROUND_VEHICLES
from fr24.proto._common_pb2 import HELICOPTERS as HELICOPTERS
from fr24.proto._common_pb2 import ISS as ISS
from fr24.proto._common_pb2 import Icon as Icon
from fr24.proto._common_pb2 import ImageInfo as ImageInfo
from fr24.proto._common_pb2 import LIFEGUARD_MEDICAL_EMERGENCY as LIFEGUARD_MEDICAL_EMERGENCY
from fr24.proto._common_pb2 import LIGHTER_THAN_AIR as LIGHTER_THAN_AIR
from fr24.proto._common_pb2 import LIVE as LIVE
from fr24.proto._common_pb2 import LJ60 as LJ60
from fr24.proto._common_pb2 import MD11 as MD11
from fr24.proto._common_pb2 import MILITARY_AND_GOVERNMENT as MILITARY_AND_GOVERNMENT
from fr24.proto._common_pb2 import MINIMUM_FUEL as MINIMUM_FUEL
from fr24.proto._common_pb2 import MLAT as MLAT
from fr24.proto._common_pb2 import NONE as NONE
from fr24.proto._common_pb2 import NON_CATEGORIZED as NON_CATEGORIZED
from fr24.proto._common_pb2 import NORMAL as NORMAL
from fr24.proto._common_pb2 import NOT_AVAILABLE as NOT_AVAILABLE
from fr24.proto._common_pb2 import NOT_VISIBLE as NOT_VISIBLE
from fr24.proto._common_pb2 import NO_COMMUNICATIONS as NO_COMMUNICATIONS
from fr24.proto._common_pb2 import NO_EMERGENCY as NO_EMERGENCY
from fr24.proto._common_pb2 import ON_GROUND as ON_GROUND
from fr24.proto._common_pb2 import OTHER_DATA_SOURCE as OTHER_DATA_SOURCE
from fr24.proto._common_pb2 import OTHER_SERVICE as OTHER_SERVICE
from fr24.proto._common_pb2 import PARTIALLY_VISIBLE as PARTIALLY_VISIBLE
from fr24.proto._common_pb2 import PASSENGER as PASSENGER
from fr24.proto._common_pb2 import Point as Point
from fr24.proto._common_pb2 import PositionBuffer as PositionBuffer
from fr24.proto._common_pb2 import Q300 as Q300
from fr24.proto._common_pb2 import RED as RED
from fr24.proto._common_pb2 import RESERVED as RESERVED
from fr24.proto._common_pb2 import RJ85 as RJ85
from fr24.proto._common_pb2 import RadarHistoryRecord as RadarHistoryRecord
from fr24.proto._common_pb2 import RecentPosition as RecentPosition
from fr24.proto._common_pb2 import RestrictionVisibility as RestrictionVisibility
from fr24.proto._common_pb2 import Route as Route
from fr24.proto._common_pb2 import SAT as SAT
from fr24.proto._common_pb2 import SATELLITE as SATELLITE
from fr24.proto._common_pb2 import SI2 as SI2
from fr24.proto._common_pb2 import SLEI as SLEI
from fr24.proto._common_pb2 import SPIDERTRACKS as SPIDERTRACKS
from fr24.proto._common_pb2 import Schedule as Schedule
from fr24.proto._common_pb2 import ScheduleInfo as ScheduleInfo
from fr24.proto._common_pb2 import Service as Service
from fr24.proto._common_pb2 import SourceStats as SourceStats
from fr24.proto._common_pb2 import Stats as Stats
from fr24.proto._common_pb2 import Status as Status
from fr24.proto._common_pb2 import Tick as Tick
from fr24.proto._common_pb2 import TrafficType as TrafficType
from fr24.proto._common_pb2 import TrailPoint as TrailPoint
from fr24.proto._common_pb2 import UAT as UAT
from fr24.proto._common_pb2 import UNKNOWN as UNKNOWN
from fr24.proto._common_pb2 import UNLAWFUL_INTERFERENCE as UNLAWFUL_INTERFERENCE
from fr24.proto._common_pb2 import YELLOW as YELLOW
from fr24.proto._fetch_search_index_pb2 import FetchSearchIndexRequest as FetchSearchIndexRequest
from fr24.proto._fetch_search_index_pb2 import FetchSearchIndexResponse as FetchSearchIndexResponse
from fr24.proto._fetch_search_index_pb2 import FlightSearchData as FlightSearchData
from fr24.proto._flight_details_pb2 import FlightDetailsRequest as FlightDetailsRequest
from fr24.proto._flight_details_pb2 import FlightDetailsResponse as FlightDetailsResponse
from fr24.proto._flight_details_pb2 import FlightProgress as FlightProgress_1
from fr24.proto._follow_flight_pb2 import FollowFlightRequest as FollowFlightRequest
from fr24.proto._follow_flight_pb2 import FollowFlightResponse as FollowFlightResponse
from fr24.proto._health_pb2 import Ping as Ping
from fr24.proto._health_pb2 import Pong as Pong
from fr24.proto._historic_trail_pb2 import HistoricTrailRequest as HistoricTrailRequest
from fr24.proto._historic_trail_pb2 import HistoricTrailResponse as HistoricTrailResponse
from fr24.proto._live_feed_pb2 import AirlineFilter as AirlineFilter
from fr24.proto._live_feed_pb2 import AirlineFilterType as AirlineFilterType
from fr24.proto._live_feed_pb2 import AirportFilter as AirportFilter
from fr24.proto._live_feed_pb2 import AirportFilterType as AirportFilterType
from fr24.proto._live_feed_pb2 import BOTH as BOTH
from fr24.proto._live_feed_pb2 import Filter as Filter
from fr24.proto._live_feed_pb2 import INBOUND as INBOUND
from fr24.proto._live_feed_pb2 import Interval as Interval
from fr24.proto._live_feed_pb2 import LiveFeedRequest as LiveFeedRequest
from fr24.proto._live_feed_pb2 import LiveFeedResponse as LiveFeedResponse
from fr24.proto._live_feed_pb2 import LocationBoundaries as LocationBoundaries
from fr24.proto._live_feed_pb2 import OPERATED_BY as OPERATED_BY
from fr24.proto._live_feed_pb2 import OUTBOUND as OUTBOUND
from fr24.proto._live_feed_pb2 import PAINTED_AS as PAINTED_AS
from fr24.proto._live_feed_pb2 import PlaybackRequest as PlaybackRequest
from fr24.proto._live_feed_pb2 import PlaybackResponse as PlaybackResponse
from fr24.proto._live_feed_pb2 import VisibilitySettings as VisibilitySettings
from fr24.proto._live_flight_status_pb2 import LiveFlightStatus as LiveFlightStatus
from fr24.proto._live_flight_status_pb2 import LiveFlightStatusData as LiveFlightStatusData
from fr24.proto._live_flight_status_pb2 import LiveFlightsStatusRequest as LiveFlightsStatusRequest
from fr24.proto._live_flight_status_pb2 import LiveFlightsStatusResponse as LiveFlightsStatusResponse
from fr24.proto._live_trail_pb2 import LiveTrailRequest as LiveTrailRequest
from fr24.proto._live_trail_pb2 import LiveTrailResponse as LiveTrailResponse
from fr24.proto._nearest_flights_pb2 import Geolocation as Geolocation
from fr24.proto._nearest_flights_pb2 import NearbyFlight as NearbyFlight
from fr24.proto._nearest_flights_pb2 import NearestFlightsRequest as NearestFlightsRequest
from fr24.proto._nearest_flights_pb2 import NearestFlightsResponse as NearestFlightsResponse
from fr24.proto._playback_flight_pb2 import PlaybackFlightRequest as PlaybackFlightRequest
from fr24.proto._playback_flight_pb2 import PlaybackFlightResponse as PlaybackFlightResponse
from fr24.proto._top_flights_pb2 import FollowedFlight as FollowedFlight
from fr24.proto._top_flights_pb2 import TopFlightsRequest as TopFlightsRequest
from fr24.proto._top_flights_pb2 import TopFlightsResponse as TopFlightsResponse
from google.protobuf import descriptor as descriptor

DESCRIPTOR: descriptor.FileDescriptor



# ruff: noqa
# mypy: disable-error-code="top-level-await, no-redef"
# %%
# --8<-- [start:script0]
from fr24.types.json import Find
from fr24 import FR24

import polars as pl


async def my_find() -> Find:
    async with FR24() as fr24:
        result = await fr24.find.fetch(query="paris")
        return result.to_dict()


results = await my_find()
df = pl.json_normalize(results["results"])
print(df)
# --8<-- [end:script0]
# %%
# --8<-- [start:script1]
from fr24.types.json import Find
from fr24.json import find, FindParams
from fr24.proto.headers import get_grpc_headers

import polars as pl
import httpx


async def my_find() -> Find:
    headers = httpx.Headers(get_grpc_headers(auth=None))
    async with httpx.AsyncClient() as client:
        response = await find(
            client, FindParams(query="paris"), headers=headers, auth=None
        )
        response.raise_for_status()
        results = response.json()
        return results  # type: ignore


results = await my_find()
df = pl.json_normalize(results["results"])
print(df)
# --8<-- [end:script1]
# %%
"""
# --8<-- [start:df0]
shape: (9, 7)
┌─────┬───────────────────────────────┬─────────┬──────────┬────────────┬────────────┬─────────────┐
│ id  ┆ label                         ┆ type    ┆ match    ┆ detail.lat ┆ detail.lon ┆ detail.size │
│ --- ┆ ---                           ┆ ---     ┆ ---      ┆ ---        ┆ ---        ┆ ---         │
│ str ┆ str                           ┆ str     ┆ str      ┆ f64        ┆ f64        ┆ i64         │
╞═════╪═══════════════════════════════╪═════════╪══════════╪════════════╪════════════╪═════════════╡
│ CDG ┆ Paris Charles de Gaulle       ┆ airport ┆ begins   ┆ 49.012516  ┆ 2.555752   ┆ 201280      │
│     ┆ Airpor…                       ┆         ┆          ┆            ┆            ┆             │
│ ORY ┆ Paris Orly Airport (ORY /     ┆ airport ┆ begins   ┆ 48.723331  ┆ 2.379444   ┆ 84510       │
│     ┆ LFPO…                         ┆         ┆          ┆            ┆            ┆             │
│ LBG ┆ Paris Le Bourget Airport (LBG ┆ airport ┆ begins   ┆ 48.958801  ┆ 2.4336     ┆ 24465       │
│     ┆ …                             ┆         ┆          ┆            ┆            ┆             │
│ BVA ┆ Paris Beauvais-Tille Airport  ┆ airport ┆ begins   ┆ 49.453465  ┆ 2.115138   ┆ 15939       │
│     ┆ (…                            ┆         ┆          ┆            ┆            ┆             │
│ VIY ┆ Paris Villacoublay Velizy Air ┆ airport ┆ begins   ┆ 48.774399  ┆ 2.20154    ┆ 1038        │
│     ┆ …                             ┆         ┆          ┆            ┆            ┆             │
│ PRX ┆ Paris Cox Field (PRX / KPRX)  ┆ airport ┆ begins   ┆ 33.636665  ┆ -95.450279 ┆ 953         │
│ XCR ┆ Paris Vatry Chalons Airport   ┆ airport ┆ begins   ┆ 48.759998  ┆ 4.2        ┆ 710         │
│     ┆ (X…                           ┆         ┆          ┆            ┆            ┆             │
│ PHT ┆ Paris Henry County Airport    ┆ airport ┆ begins   ┆ 36.336384  ┆ -88.38427  ┆ 199         │
│     ┆ (PH…                          ┆         ┆          ┆            ┆            ┆             │
│ OPL ┆ Opelousas St. Landry Parish   ┆ airport ┆ contains ┆ 30.555695  ┆ -92.098915 ┆ 194         │
│     ┆ Ai…                           ┆         ┆          ┆            ┆            ┆             │
└─────┴───────────────────────────────┴─────────┴──────────┴────────────┴────────────┴─────────────┘
# --8<-- [end:df0]
"""



from __future__ import annotations

import email.utils
import logging
import re
import sys
import time
from dataclasses import dataclass, field
from datetime import datetime, timezone
from pathlib import Path
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterator,
    Literal,
    NamedTuple,
    Protocol,
    Sequence,
    TypeVar,
    Union,
    overload,
)

from typing_extensions import (
    assert_never,
    dataclass_transform,
    runtime_checkable,
)

dataclass_opts: dict[str, bool] = {}
if sys.version_info >= (3, 10):
    dataclass_opts["slots"] = True

if TYPE_CHECKING:
    from typing import NoReturn

    import httpx
    import polars as pl
    from typing_extensions import TypeAlias

    from .types import (
        IntFlightId,
        IntoFlightId,
        IntoTimestamp,
        StrFlightIdHex,
    )
    from .types.cache import TabularFileFmt
    from .types.isqx import TimestampS

    _D = TypeVar("_D")

    @dataclass_transform(frozen_default=True)
    def dataclass_frozen(cls: type[_D]) -> type[_D]: ...
else:
    dataclass_frozen = dataclass(**dataclass_opts)


class OptionalDependencyError(ImportError):
    """Raised when an optional dependency is required but unavailable."""


def raise_missing_polars(exc: ImportError) -> NoReturn:
    raise OptionalDependencyError(
        "polars is required for dataframe and table features. "
        "Install it with `pip install fr24[polars]` or `pip install polars`."
    ) from exc


logger = logging.getLogger(__name__)

DEFAULT_HEADERS = {
    "User-Agent": (
        "Mozilla/5.0 (X11; Linux x86_64; rv:148.0) Gecko/20100101 Firefox/148.0"
    ),
    "Accept": "text/html,application/xhtml+xml,application/xml;"
    "q=0.9,image/avif,image/webp,*/*;q=0.8",
    "Accept-Language": "en-US,en;q=0.5",
    "Accept-Encoding": "gzip, deflate, zstd",  # prior to v0.2.0, this was `br`
    "Origin": "https://www.flightradar24.com",
    "Connection": "keep-alive",
    "Referer": "https://www.flightradar24.com/",
    "Sec-Fetch-Dest": "empty",
    "Sec-Fetch-Mode": "cors",
    "Sec-Fetch-Site": "same-site",
    "Sec-GPC": "1",
    "TE": "trailers",
}


@overload
def to_unix_timestamp(
    timestamp: IntoTimestamp | Literal["now"] | str,
    *,
    format: str | None = None,
) -> TimestampS[int] | Literal["now"]: ...


@overload
def to_unix_timestamp(
    timestamp: None, *, format: str | None = None
) -> None: ...


def to_unix_timestamp(
    timestamp: IntoTimestamp | str | Literal["now"] | None,
    *,
    format: str | None = None,
) -> TimestampS[int] | Literal["now"] | None:
    """Casts datetime-like object to a Unix timestamp in **integer** seconds.

    :param timestamp: Datetime-like object, such as `2025-10-13T10:30:00+05:30`,
        `2025/10/13`, `01-02-2025` etc.
        If you wish to parse non-ISO8601 strings (e.g. named/abbreviated month
        `Mar` or `March`, timezone abbreviations `EST` etc.), please use the
        `dateutil` package.
    :param format: Format to use for conversion: see the [`chrono` documentation](https://docs.rs/chrono/latest/chrono/format/strftime/index.html).
        If `None` (default), the format is inferred against a
        [known sequence of pattern groups](https://github.com/pola-rs/polars/blob/a765e92556e0a6f14760dc90612bf1511dcee864/crates/polars-time/src/chunkedarray/string/patterns.rs).
    """
    if isinstance(timestamp, str):
        if timestamp == "now":
            # the caller is responsible for calling `get_current_timestamp()`,
            # which can happen at a later time during the API `fetch()` call
            return "now"
        # try to interpret string as integer timestamp in seconds first
        try:
            return to_unix_timestamp(int(timestamp))
        except ValueError:
            pass
        # otherwise assume it's a format that `polars` understands
        try:
            import polars as pl
        except ImportError as exc:
            raise_missing_polars(exc)

        try:
            dt: datetime = (
                pl.Series(values=(timestamp,))
                .str.to_datetime(format=format)
                .item(0)
            )
        except pl.exceptions.ComputeError:
            raise ValueError(
                f"invalid datetime string: {timestamp!r}\n"
                "help: must be ISO8601-like."
            )
        return int(dt.timestamp())
    if isinstance(timestamp, datetime):
        return int(timestamp.timestamp())
    if isinstance(timestamp, int):
        if timestamp > 4102462800:  # 2100-01-01
            raise ValueError(
                f"timestamp {timestamp} is too large\n"
                "help: should be in seconds, not milliseconds"
            )
        return timestamp
    return None


def get_current_timestamp() -> TimestampS[int]:
    """Returns the current Unix timestamp in seconds."""
    return int(time.time())


def parse_server_timestamp(
    response: httpx.Response,
) -> TimestampS[int] | None:
    server_date: str = response.headers.get("date")
    if server_date is not None:
        return int(email.utils.parsedate_to_datetime(server_date).timestamp())
    return None


def to_flight_id(flight_id: IntoFlightId) -> IntFlightId:
    if isinstance(flight_id, (str, bytes)):
        return int(flight_id, 16)
    return flight_id


def to_flight_id_hex(flight_id: IntoFlightId) -> StrFlightIdHex:
    """Converts flight ID to a hex string."""
    if isinstance(flight_id, str):
        return flight_id.lower().removeprefix("0x")
    if isinstance(flight_id, bytes):
        flight_id = flight_id.decode()
    return f"{flight_id:x}"


@dataclass_frozen
class BarePath:
    path: Path


FileLike = Union[str, Path, IO[bytes], BarePath]


def format_bare_path(path: BarePath, format: TabularFileFmt) -> Path:
    if format == "parquet":
        suffix = ".parquet"
    elif format == "csv":
        suffix = ".csv"
    else:
        raise ValueError(f"unsupported format: `{format}`")
    return path.path.with_suffix(suffix)


FileExistsBehaviour: TypeAlias = Literal["backup", "error", "overwrite"]


def _handle_existing_file(
    path: Path, when_file_exists: FileExistsBehaviour
) -> None:
    if not path.exists():
        return
    if when_file_exists == "backup":
        timestamp = datetime.now(tz=timezone.utc).isoformat()
        backup_path = path.with_suffix(f"{path.suffix}.bkp.{timestamp}")
        path.rename(backup_path)
        logger.info(
            f"file `{path}` already exists, moved it to `{backup_path}`"
        )
    elif when_file_exists == "error":
        raise FileExistsError(f"refused to overwrite existing file at `{path}`")
    elif when_file_exists == "overwrite":
        pass
    else:
        assert_never(when_file_exists)


def write_table(
    result: SupportsToPolars,
    file: FileLike,
    *,
    format: TabularFileFmt = "parquet",
    when_file_exists: FileExistsBehaviour = "backup",
    **kwargs: Any,
) -> None:
    """Writes the table as the specified format via polars.

    :param file: File path or writable file-like object. The path will be given
        an appropriate suffix if it is a [BarePath][fr24.utils.BarePath].
    """

    if isinstance(file, BarePath):
        file = format_bare_path(file, format)
    if isinstance(file, str):
        file = Path(file)
    if isinstance(file, Path):
        _handle_existing_file(file, when_file_exists)
        file.parent.mkdir(parents=True, exist_ok=True)

    data = result.to_polars()
    if format == "parquet":
        # NOTE: saving metadata with polars is not yet implemented
        # this means that useful unstructured metadata (e.g. flight details)
        # cannot be saved without extra pyarrow dependency
        # https://github.com/pola-rs/polars/issues/5117
        data.write_parquet(file, **kwargs)
    elif format == "csv":
        data.write_csv(file, **kwargs)
    else:
        raise ValueError(f"unsupported format: `{format}`")
    logger.info(f"wrote {data.height} rows to `{file}`")


def scan_table(
    file: FileLike,
    *,
    format: TabularFileFmt = "parquet",
    schema: dict[str, pl.DataType] | None = None,
) -> pl.LazyFrame:
    """
    Reads the table as the specified format via polars.

    :param file: File path or readable file-like object. The path will be given
        an appropriate suffix if it is a [BarePath][fr24.utils.BarePath].
    :param schema: The schema to enforce when reading the table. For CSV files,
        this should be specified to properly parse datetimes from strings.
    """
    try:
        import polars as pl
    except ImportError as exc:
        raise_missing_polars(exc)

    if isinstance(file, BarePath):
        file = format_bare_path(file, format)

    if format == "parquet":
        return pl.scan_parquet(file, schema=schema)
    elif format == "csv":
        return pl.scan_csv(file, schema=schema)
    else:
        raise ValueError(f"unsupported format: `{format}`")


#
# important traits to perform conversions to other formats
# Into<TypedDict | google.protobuf.Message | pl.DataFrame>
# subclassing is not necessary, but recommended for static type checking
#

DictT_co = TypeVar("DictT_co", covariant=True)
"""The dictionary representation of an object, e.g. `TypedDict`."""


@runtime_checkable
class SupportsToDict(Protocol[DictT_co]):
    def to_dict(self) -> DictT_co:
        """Converts the object into a dictionary."""
        ...


@runtime_checkable
class SupportsToPolars(Protocol):
    def to_polars(self) -> pl.DataFrame:
        """Converts the object into a polars dataframe."""
        ...


# instead of raising exceptions:
# - in `fr24.json`: return a result type that stores the response body or any
#   errors
# - in `fr24.grpc`: return a result type that stores the parsed protocol buffer
#   or any gRPC errors

T = TypeVar("T")
E = TypeVar("E")


@dataclass_frozen
class Ok(Generic[T]):
    _value: T

    def ok(self) -> T:
        return self._value

    def err(self) -> None:
        return None

    def is_ok(self) -> Literal[True]:
        return True

    def is_err(self) -> Literal[False]:
        return False

    def unwrap(self) -> T:
        return self._value


@dataclass_frozen
class Err(Generic[E]):
    _value: E

    def ok(self) -> None:
        return None

    def err(self) -> E:
        return self._value

    def is_ok(self) -> Literal[False]:
        return False

    def is_err(self) -> Literal[True]:
        return True

    def unwrap(self) -> NoReturn:
        raise UnwrapError(
            self, f"called `Result.unwrap()` on errored result: {self._value!r}"
        )


class UnwrapError(Exception):
    def __init__(self, err: Any, message: str) -> None:
        self._err = err
        super().__init__(message)

    @property
    def err(self) -> Any:
        return self._err


Result: TypeAlias = Union[Ok[T], Err[E]]
"""A type that represents either success (`Ok`) or failure (`Err`)."""

M = TypeVar("M", bound=Callable[..., Any])


def static_check_signature(dataclass: Any) -> Callable[[M], M]:
    '''Marker to signal the static analyser that the decorated method or
    function should be checked against the signature of the given dataclass.

    Many of our low level APIs use 1) parameters stored in a single flat
    dataclass, and 2) functions that take in this dataclass:
    ```py
    @dataclass
    class LiveFeedParams:
        bbox_south: float
        """Latitude, minimum, degrees"""
        bbox_north: float
        bbox_west: float
        bbox_east: float
        flight_id: int
        # ... other params ...

    def live_feed(params: LiveFeedParams, ...):
        ...
    ```
    but service methods and CLIs expect a flat signature:
    ```py
    from fr24.utils import static_check_signature

    @static_check_signature(LiveFeedParams)
    def live_feed_cli(
        bbox_south: float,
        bbox_north: float,
        bbox_west: float,
        bbox_east: float,
        flight_id: int
        # ... other params ...
    ):
        """:param bbox_south: The southern latitude of the bounding box.
        ..."""
        ...
    ```
    `./scripts/check_signature.py` uses static analysis to:

    1. scan for all uses of the decorator
    2. for each member of the dataclass, collect all docstrings, members names
       and types
    3. for each argument of the decorated method, parse the sphinx docstring,
       names and types
    4. compare them.
    '''

    def decorator(method: M) -> M:
        return method

    return decorator


# for use in cli
_SPHINX_PARAM = re.compile(r"\s*:param\s+(?P<name>\w+):\s*(?P<doc>.*)")


class ParamDetail(NamedTuple):
    name: str
    description: str


@dataclass
class SphinxParser:
    """A poor man's parser for Sphinx docstrings and its `:param:` directive."""

    text: str
    name: str | None = None
    doc_lines: list[str] = field(default_factory=list)

    def __iter__(self) -> Iterator[str | ParamDetail]:
        for line in self.text.splitlines():
            if match := _SPHINX_PARAM.match(line):
                if self.name is not None:
                    yield ParamDetail(
                        self.name, _join_doc_lines(self.doc_lines)
                    )
                elif self.doc_lines:  # header
                    yield _join_doc_lines(self.doc_lines)
                self.name = match.group("name")
                self.doc_lines = [match.group("doc")]
            else:
                self.doc_lines.append(line)
        if self.name:
            yield ParamDetail(self.name, _join_doc_lines(self.doc_lines))
        else:
            yield _join_doc_lines(self.doc_lines)


def _join_doc_lines(doc_frags: Sequence[str]) -> str:
    return " ".join(line.strip() for line in doc_frags)



# ruff: noqa
# mypy: disable-error-code="top-level-await, no-redef"
# %%
# --8<-- [start:script0]
import httpx
from fr24.grpc import FlightDetailsParams, flight_details
from fr24.proto.v1_pb2 import FlightDetailsResponse
from fr24.proto import parse_data
from fr24.proto.headers import get_grpc_headers


async def flight_details_data() -> FlightDetailsResponse:
    headers = httpx.Headers(get_grpc_headers(auth=None))
    async with httpx.AsyncClient() as client:
        params = FlightDetailsParams(flight_id=0x3C500FDB)
        response = await flight_details(client, params, headers)
        return parse_data(response.content, FlightDetailsResponse).unwrap()


data = await flight_details_data()
data
# --8<-- [end:script0]
# %%
"""
# --8<-- [start:output0]
aircraft_info {
  icao_address: 7866977
  reg: "B-KQM"
  type: "B77W"
  icon: B777
  full_description: "Boeing 777-367(ER)"
  images_list {
    url: "https://www.jetphotos.com/photo/11825817"
    copyright: "ZSHC_Linzx"
    thumbnail: "https://cdn.jetphotos.com/200/5/1291679_1756473618_tb.jpg"
    medium: "https://cdn.jetphotos.com/400/5/1291679_1756473618.jpg"
    large: "https://cdn.jetphotos.com/640/5/1291679_1756473618.jpg"
  }
  images_list {
    url: "https://www.jetphotos.com/photo/11813969"
    copyright: "CAN-Eric"
    thumbnail: "https://cdn.jetphotos.com/200/5/967682_1755532162_tb.jpg"
    medium: "https://cdn.jetphotos.com/400/5/967682_1755532162.jpg"
    large: "https://cdn.jetphotos.com/640/5/967682_1755532162.jpg"
  }
  images_list {
    url: "https://www.jetphotos.com/photo/11803561"
    copyright: "cc12214"
    thumbnail: "https://cdn.jetphotos.com/200/6/585221_1754572527_tb.jpg"
    medium: "https://cdn.jetphotos.com/400/6/585221_1754572527.jpg"
    large: "https://cdn.jetphotos.com/640/6/585221_1754572527.jpg"
  }
  msn_available: true
  age_available: true
  registered_owners: "Cathay Pacific (Oneworld Livery)"
  is_country_of_reg_available: true
}
schedule_info {
  flight_number: "CX251"
  operated_by_id: 57
  painted_as_id: 57
  origin_id: 1366
  destination_id: 1942
  scheduled_departure: 1758465600
  scheduled_arrival: 1758516000
  actual_departure: 1758466901
  arr_terminal: "3"
}
flight_progress {
  traversed_distance: 13728
  remaining_distance: 9652855
  elapsed_time: 156
  remaining_time: 51708
  eta: 1758518765
  great_circle_distance: 9647791
  mean_flight_time: 48161
  flight_stage: ASCENDING
  delay_status: RED
}
flight_info {
  flightid: 1011879899
  lat: 22.3374939
  lon: 114.046074
  track: 99
  alt: 4975
  speed: 205
  timestamp_ms: 1758467055960
  callsign: "CPA251"
  ems_availability {
    qnh_availability: true
    amcp_availability: true
    oat_availability: true
    ias_availability: true
    tas_availability: true
    mach_availability: true
    agps_availability: true
    agpsdiff_availability: true
    wind_dir_availability: true
    wind_speed_availability: true
    rs_availability: true
  }
  squawk_availability: true
  vspeed_availability: true
  airspace_availability: true
  server_time_ms: 1758467057331
}
flight_plan {
}
flight_trail_list {
  snapshot_id: 1758465991
  lat: 22.3132782
  lon: 113.934616
  heading: 70
}
...
flight_trail_list {
  snapshot_id: 1758467049
  lat: 22.3383179
  lon: 114.039337
  altitude: 4800
  spd: 206
  heading: 91
}
position_buffer {
  recent_positions_list {
    delta_lat: -45
    delta_lon: 274
    delta_ms: 2710
  }
  recent_positions_list {
    delta_lat: -50
    delta_lon: 324
    delta_ms: 3180
  }
  recent_positions_list {
    delta_lat: -78
    delta_lon: 585
    delta_ms: 5720
  }
  recent_positions_list {
    delta_lat: -86
    delta_lon: 778
    delta_ms: 7570
  }
  recent_positions_list {
    delta_lat: -92
    delta_lon: 976
    delta_ms: 9540
  }
  recent_positions_list {
    delta_lat: -91
    delta_lon: 1073
    delta_ms: 10500
  }
}
# --8<-- [end:output0]
"""



from __future__ import annotations

import importlib.util
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Literal,
    TypedDict,
    # NOTE: using `Union[X, Y]` instead of X | Y because get_type_hints
    # with extras on py39 fails
    Union,
    get_type_hints,
)

if TYPE_CHECKING:
    import polars as pl

from . import IntFlightId
from .isqx import (
    AltitudeFt,
    BankAngleDeg,
    BearingDeg,
    DistanceM,
    DurationS,
    GeometricAltitudeM,
    GroundSpeedKt,
    GroundTrackDeg,
    IndicatedAirSpeedKt,
    LatitudeDeg,
    LongitudeDeg,
    MachTimes1K,
    PressureAltimeterHpa,
    StaticTemperatureC,
    TimestampMs,
    TimestampS,
    TrueAirSpeedKt,
    VerticalSpeedFpm,
    WindSpeedKt,
)

# NOTE: Parquet does not support timestamp in seconds:
# https://github.com/apache/parquet-format/blob/master/LogicalTypes.md#timestamp
# using u32 or timestamp_ms for now


@dataclass(frozen=True)
class DType:
    """A newtype for annotating types in TypedDicts."""

    type_: pl.DataType

    def __post_init__(self) -> None:
        if importlib.util.find_spec("polars") is None:
            return
        import polars as pl

        assert isinstance(self.type_, pl.DataType), (
            f"found `{type(self.type_)=}`"
        )


def to_schema(obj: type[Any]) -> dict[str, pl.DataType] | None:
    """Generate a polars schema from a TypedDict.

    :param obj: A TypedDict with types annotated with `DType`.
    :returns: None if polars is not installed
    """
    if importlib.util.find_spec("polars") is None:
        return None
    import polars as pl

    schema = {}
    hints = get_type_hints(obj, include_extras=True, localns={"pl": pl})
    for field_name, type_ in hints.items():
        metadata_list = getattr(type_, "__metadata__", None)
        assert metadata_list is not None, (
            f"expected {field_name=} to be `Annotated`, found `{type_}`"
        )
        dtype = None
        for metadata in metadata_list:
            if isinstance(metadata, DType):
                if dtype is not None:
                    raise ValueError(
                        f"multiple `DType` annotations found for {field_name=}"
                    )
                dtype = metadata.type_
        if dtype is None:
            raise ValueError(f"no `DType` found for {field_name=}")
        schema[field_name] = dtype
    return schema


# NOTE: using `Union[X, Y]` instead of X | Y because get_type_hints with extras
# on py39 fails
class FlightListRecord(TypedDict):
    flight_id: Annotated[Union[IntFlightId, None], DType(pl.UInt64())]
    number: Annotated[Union[str, None], DType(pl.String())]
    callsign: Annotated[Union[str, None], DType(pl.String())]
    icao24: Annotated[Union[int, None], DType(pl.UInt32())]
    registration: Annotated[Union[str, None], DType(pl.String())]
    typecode: Annotated[str, DType(pl.String())]
    origin: Annotated[Union[str, None], DType(pl.String())]
    destination: Annotated[Union[str, None], DType(pl.String())]
    status: Annotated[Union[str, None], DType(pl.String())]
    STOD: Annotated[
        Union[TimestampMs[int], None], DType(pl.Datetime("ms", time_zone="UTC"))
    ]
    ETOD: Annotated[
        Union[TimestampMs[int], None], DType(pl.Datetime("ms", time_zone="UTC"))
    ]
    ATOD: Annotated[
        Union[TimestampMs[int], None], DType(pl.Datetime("ms", time_zone="UTC"))
    ]
    STOA: Annotated[
        Union[TimestampMs[int], None], DType(pl.Datetime("ms", time_zone="UTC"))
    ]
    ETOA: Annotated[
        Union[TimestampMs[int], None], DType(pl.Datetime("ms", time_zone="UTC"))
    ]
    ATOA: Annotated[
        Union[TimestampMs[int], None], DType(pl.Datetime("ms", time_zone="UTC"))
    ]


flight_list_schema = to_schema(FlightListRecord)


class PlaybackTrackEMSRecord(TypedDict):
    timestamp: Annotated[TimestampS[int], DType(pl.UInt32())]
    ias: Annotated[Union[IndicatedAirSpeedKt[int], None], DType(pl.Int16())]
    tas: Annotated[Union[TrueAirSpeedKt[int], None], DType(pl.Int16())]
    mach: Annotated[Union[MachTimes1K[int], None], DType(pl.Int16())]
    mcp: Annotated[Union[AltitudeFt[int], None], DType(pl.Int32())]
    fms: Annotated[Union[AltitudeFt[int], None], DType(pl.Int32())]
    autopilot: Annotated[Union[int, None], DType(pl.Int8())]
    oat: Annotated[Union[StaticTemperatureC[int], None], DType(pl.Int8())]
    track: Annotated[Union[GroundTrackDeg[float], None], DType(pl.Float32())]
    roll: Annotated[Union[BankAngleDeg[float], None], DType(pl.Float32())]
    qnh: Annotated[Union[PressureAltimeterHpa[int], None], DType(pl.UInt16())]
    wind_dir: Annotated[Union[BearingDeg[int], None], DType(pl.Int16())]
    wind_speed: Annotated[Union[WindSpeedKt[int], None], DType(pl.Int16())]
    precision: Annotated[Union[int, None], DType(pl.UInt8())]
    altitude_gps: Annotated[
        Union[GeometricAltitudeM[int], None], DType(pl.Int32())
    ]
    emergency: Annotated[Union[int, None], DType(pl.UInt8())]
    tcas_acas: Annotated[Union[int, None], DType(pl.UInt8())]
    heading: Annotated[Union[BearingDeg[int], None], DType(pl.UInt16())]


playback_track_ems_schema = to_schema(PlaybackTrackEMSRecord)


class PlaybackTrackRecord(TypedDict):
    timestamp: Annotated[TimestampS[int], DType(pl.UInt32())]
    latitude: Annotated[LatitudeDeg[float], DType(pl.Float32())]
    longitude: Annotated[LongitudeDeg[float], DType(pl.Float32())]
    altitude: Annotated[AltitudeFt[int], DType(pl.Int32())]
    ground_speed: Annotated[GroundSpeedKt[int], DType(pl.Int16())]
    vertical_speed: Annotated[
        Union[VerticalSpeedFpm[int], None], DType(pl.Int16())
    ]
    track: Annotated[GroundTrackDeg[int], DType(pl.Int16())]
    squawk: Annotated[int, DType(pl.UInt16())]
    ems: Annotated[
        Union[None, PlaybackTrackEMSRecord],
        DType(pl.Struct(playback_track_ems_schema)),
    ]


playback_track_schema = to_schema(PlaybackTrackRecord)


class RecentPositionRecord(TypedDict):
    delta_lat: Annotated[int, DType(pl.Int32())]
    delta_lon: Annotated[int, DType(pl.Int32())]
    delta_ms: Annotated[int, DType(pl.UInt32())]


position_buffer_struct_schema = to_schema(RecentPositionRecord)


class FlightRecord(TypedDict):
    timestamp: Annotated[
        TimestampMs[int], DType(pl.Datetime("ms", time_zone="UTC"))
    ]  # prior to v0.2.0 this was u32 (seconds), now i64 (milliseconds)
    flightid: Annotated[IntFlightId, DType(pl.UInt32())]
    latitude: Annotated[LatitudeDeg[float], DType(pl.Float32())]
    longitude: Annotated[LongitudeDeg[float], DType(pl.Float32())]
    track: Annotated[GroundTrackDeg[int], DType(pl.UInt16())]
    altitude: Annotated[AltitudeFt[int], DType(pl.Int32())]
    ground_speed: Annotated[GroundSpeedKt[int], DType(pl.Int16())]
    on_ground: Annotated[bool, DType(pl.Boolean())]
    callsign: Annotated[str, DType(pl.String())]
    source: Annotated[int, DType(pl.UInt8())]
    registration: Annotated[str, DType(pl.String())]
    origin: Annotated[str, DType(pl.String())]
    destination: Annotated[str, DType(pl.String())]
    typecode: Annotated[str, DType(pl.String())]
    eta: Annotated[TimestampS[int], DType(pl.UInt32())]
    squawk: Annotated[int, DType(pl.UInt16())]
    vertical_speed: Annotated[
        Union[VerticalSpeedFpm[int], None], DType(pl.Int16())
    ]  # 64 * 9-bit + 1-bit sign
    position_buffer: Annotated[
        list[RecentPositionRecord],
        DType(pl.List(pl.Struct(position_buffer_struct_schema))),
    ]


live_feed_schema = to_schema(FlightRecord)


class NearbyFlightRecord(FlightRecord):
    distance: Annotated[DistanceM[int], DType(pl.UInt32())]


nearest_flights_schema = to_schema(NearbyFlightRecord)


class LiveFlightStatusRecord(TypedDict):
    flight_id: Annotated[IntFlightId, DType(pl.UInt32())]
    latitude: Annotated[LatitudeDeg[float], DType(pl.Float32())]
    longitude: Annotated[LongitudeDeg[float], DType(pl.Float32())]
    status: Annotated[int, DType(pl.UInt8())]
    squawk: Annotated[int, DType(pl.UInt16())]


live_flights_status_schema = to_schema(LiveFlightStatusRecord)


class TopFlightRecord(TypedDict):
    flight_id: Annotated[IntFlightId, DType(pl.UInt32())]
    live_clicks: Annotated[int, DType(pl.UInt32())]
    total_clicks: Annotated[int, DType(pl.UInt32())]
    flight_number: Annotated[str, DType(pl.String())]
    callsign: Annotated[str, DType(pl.String())]
    squawk: Annotated[int, DType(pl.UInt16())]
    from_iata: Annotated[str, DType(pl.String())]
    from_city: Annotated[str, DType(pl.String())]
    to_iata: Annotated[str, DType(pl.String())]
    to_city: Annotated[str, DType(pl.String())]
    type: Annotated[str, DType(pl.String())]
    full_description: Annotated[str, DType(pl.String())]


top_flights_schema = to_schema(TopFlightRecord)


class EMSRecord(TypedDict):
    ias: Annotated[Union[IndicatedAirSpeedKt[int], None], DType(pl.Int16())]
    tas: Annotated[Union[TrueAirSpeedKt[int], None], DType(pl.Int16())]
    mach: Annotated[Union[MachTimes1K[int], None], DType(pl.Int16())]
    mcp: Annotated[Union[AltitudeFt[int], None], DType(pl.Int32())]
    fms: Annotated[Union[AltitudeFt[int], None], DType(pl.Int32())]
    oat: Annotated[Union[StaticTemperatureC[int], None], DType(pl.Int8())]
    qnh: Annotated[Union[PressureAltimeterHpa[int], None], DType(pl.UInt16())]
    wind_dir: Annotated[Union[BearingDeg[int], None], DType(pl.Int16())]
    wind_speed: Annotated[Union[WindSpeedKt[int], None], DType(pl.Int16())]
    altitude_gps: Annotated[
        Union[GeometricAltitudeM[int], None], DType(pl.Int32())
    ]
    agpsdiff: Annotated[Union[int, None], DType(pl.Int32())]
    apflags: Annotated[Union[int, None], DType(pl.Int32())]
    rs: Annotated[Union[int, None], DType(pl.Int32())]


ems_struct_schema = to_schema(EMSRecord)


class TrailPointRecord(TypedDict):
    timestamp: Annotated[TimestampS[int], DType(pl.UInt32())]
    latitude: Annotated[LatitudeDeg[float], DType(pl.Float32())]
    longitude: Annotated[LongitudeDeg[float], DType(pl.Float32())]
    altitude: Annotated[Union[AltitudeFt[int], None], DType(pl.Int32())]
    ground_speed: Annotated[Union[GroundSpeedKt[int], None], DType(pl.Int16())]
    track: Annotated[Union[GroundTrackDeg[int], None], DType(pl.UInt16())]
    vertical_speed: Annotated[
        Union[VerticalSpeedFpm[int], None], DType(pl.Int16())
    ]
    source: Annotated[Union[int, None], DType(pl.UInt8())]


trail_point_schema = to_schema(TrailPointRecord)


class _AircraftRecord(TypedDict):
    icao_address: Annotated[int, DType(pl.UInt32())]
    reg: Annotated[str, DType(pl.String())]
    typecode: Annotated[str, DType(pl.String())]


class _FlightProgressRecord(TypedDict):
    traversed_distance: Annotated[DistanceM[int], DType(pl.UInt32())]
    remaining_distance: Annotated[DistanceM[int], DType(pl.UInt32())]
    elapsed_time: Annotated[Union[DurationS[int], None], DType(pl.UInt32())]
    remaining_time: Annotated[Union[DurationS[int], None], DType(pl.UInt32())]
    eta: Annotated[TimestampS[int], DType(pl.UInt32())]
    great_circle_distance: Annotated[DistanceM[int], DType(pl.UInt32())]
    mean_flight_time: Annotated[DurationS[int], DType(pl.UInt32())]


class _ScheduleRecord(TypedDict):
    flight_number: Annotated[str, DType(pl.String())]
    origin_id: Annotated[int, DType(pl.UInt32())]
    destination_id: Annotated[int, DType(pl.UInt32())]
    diverted_id: Annotated[int, DType(pl.UInt32())]
    scheduled_departure: Annotated[TimestampS[int], DType(pl.UInt32())]
    scheduled_arrival: Annotated[TimestampS[int], DType(pl.UInt32())]
    actual_departure: Annotated[
        Union[TimestampS[int], None], DType(pl.UInt32())
    ]
    actual_arrival: Annotated[Union[TimestampS[int], None], DType(pl.UInt32())]


class _FlightInfoRecord(TypedDict):
    timestamp_ms: Annotated[
        TimestampMs[int], DType(pl.Datetime("ms", time_zone="UTC"))
    ]  # prior to v0.2.0 this was a bare u64
    flightid: Annotated[IntFlightId, DType(pl.UInt32())]
    latitude: Annotated[LatitudeDeg[float], DType(pl.Float32())]
    longitude: Annotated[LongitudeDeg[float], DType(pl.Float32())]
    track: Annotated[Union[GroundTrackDeg[int], None], DType(pl.UInt16())]
    altitude: Annotated[AltitudeFt[int], DType(pl.Int32())]
    ground_speed: Annotated[GroundSpeedKt[int], DType(pl.Int16())]
    vertical_speed: Annotated[
        Union[VerticalSpeedFpm[int], None], DType(pl.Int16())
    ]
    on_ground: Annotated[bool, DType(pl.Boolean())]
    callsign: Annotated[str, DType(pl.String())]
    squawk: Annotated[int, DType(pl.UInt16())]
    ems: Annotated[Union[None, EMSRecord], DType(pl.Struct(ems_struct_schema))]


class _FlightTrailRecord(TypedDict):
    flight_trail_list: Annotated[
        Union[None, list[TrailPointRecord]],
        DType(pl.List(pl.Struct(trail_point_schema))),
    ]


class FlightDetailsRecord(
    _AircraftRecord,
    _ScheduleRecord,
    _FlightProgressRecord,
    _FlightInfoRecord,
    _FlightTrailRecord,
):
    position_buffer: Annotated[
        list[RecentPositionRecord],
        DType(pl.List(pl.Struct(position_buffer_struct_schema))),
    ]


flight_details_schema = to_schema(FlightDetailsRecord)


class PlaybackFlightRecord(
    _AircraftRecord,
    _ScheduleRecord,
    _FlightInfoRecord,
    _FlightTrailRecord,
):
    pass


playback_flight_schema = to_schema(PlaybackFlightRecord)
TabularFileFmt = Literal["parquet", "csv"]  # TODO: support ndjson
